--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

local drag= 0.9999999 -- 0 means full drag, 1 is no drag
local bounce= 1.2 -- must be [1,2], where 1 means full stop on impact and 2 means full rebound. but 2 seems to incite "the flubber effect" so use 1.9 as a practical upper bound
local friction= 0.05

local min_f= 0 -- of friction
local max_f= 1

local min_b= 0 -- bounce
local max_b= 0.99

local min_g= 0 -- grav
local max_g= 1

local wxmin= 0 -- these define the world bounds
local wxmax= 4096
local wymin= 0
local wymax= 4096
-- locals for fast checking with the edge of the world, Vec2 for easy rotating of points
world_max= sc.Construct("Vec2", wxmax, wymax)

--[[
collision_acoords= {}
collision_bcoords= {}
collision_ccoords= {}
collision_aindices= {}
collision_bindices= {}
collision_cindices= {}
]]

function overlap(mina, maxa, minb, maxb)
  if(mina > maxb) then
    return nil
  end
  if(maxa < minb) then
    return nil
  end
  return true
end

function point_in_rect(p, r)
  local ra= r.a
  local rb= r.b
  local xmin, xmax= minmax(ra.x, rb.x)
  if(p.x < xmin or p.x > xmax) then
    return nil
  end
  local ymin, ymax= minmax(ra.y, rb.y)
  return not (p.y < ymin or p.y > ymax)
end

function rects_overlap_aabb(ra, rb)
  local ra_pos= ra.pos
  local ra_pos_x= ra_pos.x
  local ra_pos_y= ra_pos.y
  local ra_size= ra.size
  local ra_size_x= ra_size.x
  local ra_size_y= ra_size.y
  local rb_pos= rb.pos
  local rb_pos_x= rb_pos.x
  local rb_pos_y= rb_pos.y
  local rb_size= rb.size
  local rb_size_x= rb_size.x
  local rb_size_y= rb_size.y
  local xmina, xmaxa= minmax(ra_pos_x - ra_size_x, ra_pos_x + ra_size_x)
  local xminb, xmaxb= minmax(rb_pos_x - rb_size_x, rb_pos_x + rb_size_x)
  if(not overlap(xmina, xmaxa, xminb, xmaxb)) then
    return nil
  end
  local ymina, ymaxa= minmax(ra_pos_y - ra_size_y, ra_pos_y + ra_size_y)
  local yminb, ymaxb= minmax(rb_pos_y - rb_size_y, rb_pos_y + rb_size_y)
  return overlap(ymina, ymaxa, yminb, ymaxb)
end

function rects_overlap_aabb_vs_2p(ra, rb)
  local ra_pos= ra.pos
  local ra_pos_x= ra_pos.x
  local ra_pos_y= ra_pos.y
  local ra_size= ra.size
  local ra_size_x= ra_size.x
  local ra_size_y= ra_size.y
  local rb_a= rb.a
  local rb_a_x= rb_a.x
  local rb_a_y= rb_a.y
  local rb_b= rb.b
  local rb_b_x= rb_b.x
  local rb_b_y= rb_b.y
  local xmina, xmaxa= minmax(ra_pos_x - ra_size_x, ra_pos_x + ra_size_x)
  local xminb, xmaxb= minmax(rb_a_x, rb_b_x)
  if(not overlap(xmina, xmaxa, xminb, xmaxb)) then
    return nil
  end
  local ymina, ymaxa= minmax(ra_pos_y - ra_size_y, ra_pos_y + ra_size_y)
  local yminb, ymaxb= minmax(rb_a_y, rb_b_y)
  return overlap(ymina, ymaxa, yminb, ymaxb)
end

function rects_overlap_ps(ra, rb)
--  print("Checking rects.")
--  rect_print(ra)
--  rect_print(rb)
  local ra_pos= ra.pos
  local ra_pos_x= ra_pos.x
  local ra_pos_y= ra_pos.y
  local ra_size= ra.size
  local ra_size_x= ra_size.x
  local ra_size_y= ra_size.y
  local rb_pos= rb.pos
  local rb_pos_x= rb_pos.x
  local rb_pos_y= rb_pos.y
  local rb_size= rb.size
  local rb_size_x= rb_size.x
  local rb_size_y= rb_size.y
  local xmina, xmaxa= minmax(ra_pos_x, ra_pos_x + ra_size_x)
  local xminb, xmaxb= minmax(rb_pos_x, rb_pos_x + rb_size_x)
  if(not overlap(xmina, xmaxa, xminb, xmaxb)) then
    return nil
  end
  local ymina, ymaxa= minmax(ra_pos_y, ra_pos_y + ra_size_y)
  local yminb, ymaxb= minmax(rb_pos_y, rb_pos_y + rb_size_y)
  return overlap(ymina, ymaxa, yminb, ymaxb)
end

function rects_overlap_wh(ra, rb)
--  print("Checking rects.")
--  rect_print(ra)
--  rect_print(rb)
  local ra_x= ra.x
  local ra_y= ra.y
  local rb_x= rb.x
  local rb_y= rb.y
  local xmina, xmaxa= minmax(ra_x, ra_x + ra.w)
  local xminb, xmaxb= minmax(rb_x, rb_x + rb.w)
  if(not overlap(xmina, xmaxa, xminb, xmaxb)) then
    return nil
  end
  local ymina, ymaxa= minmax(ra_y, ra_y + ra.h)
  local yminb, ymaxb= minmax(rb_y, rb_y + rb.h)
  return overlap(ymina, ymaxa, yminb, ymaxb)
end

function rects_overlap_2p(ra, rb)
--  print("Checking rects.")
--  rect_print(ra)
--  rect_print(rb)
  local ra_a= ra.a
  local ra_a_x= ra_a.x
  local ra_b= ra.b
  local ra_b_x= ra_b.x
  local rb_a= rb.a
  local rb_a_x= rb_a.x
  local rb_b= rb.b
  local rb_b_x= rb_b.x
  local xmina, xmaxa= minmax(ra_a_x, ra_b_x)
  local xminb, xmaxb= minmax(rb_a_x, rb_b_x)
  if(not overlap(xmina, xmaxa, xminb, xmaxb)) then
    return nil
  end
  local ra_a_y= ra_a.y
  local ra_b_y= ra_b.y
  local rb_a_y= rb_a.y
  local rb_b_y= rb_b.y
  local ymina, ymaxa= minmax(ra_a_y, ra_b_y)
  local yminb, ymaxb= minmax(rb_a_y, rb_b_y)
  return overlap(ymina, ymaxa, yminb, ymaxb)
end

function lines_intersect_abc(la, lb)
  -- Ax + By = C
  assert(is_real_num(la.a))
  assert(is_real_num(la.b))
  assert(is_real_num(la.c))
  assert(is_real_num(lb.a))
  assert(is_real_num(lb.b))
  assert(is_real_num(lb.c))
  local res1_a= la.a
  local res1_b= la.b
  local res1_c= la.c
  local res2_a= lb.a
  local res2_b= lb.b
  local res2_c= lb.c
  --print("res1:  "..res1_a.."x + "..res1_b.."y = "..res1_c)
  --print("res2:  "..res2_a.."x + "..res2_b.."y = "..res2_c)
  if(res1_a == 0) then
    res1_a, res2_a= res2_a, res1_a
    res1_b, res2_b= res2_b, res1_b
    res1_c, res2_c= res2_c, res1_c
    --print("Swapped res1 and res2")
  end
  if(res1_a ~= 1) then
    res1_b= res1_b / res1_a
    res1_c= res1_c / res1_a
    res1_a= res1_a / res1_a
    --print("res1:  "..res1_a.."x + "..res1_b.."y = "..res1_c)
  end
  if(res2_a ~= 0) then
    res2_b= res2_b - (res1_b * res2_a)
    res2_c= res2_c - (res1_c * res2_a)
    res2_a= res2_a - (res1_a * res2_a)
    --print("res2:  "..res2_a.."x + "..res2_b.."y = "..res2_c)
  end
  if(res2_b ~= 1) then
    res2_a= res2_a / res2_b
    res2_c= res2_c / res2_b
    res2_b= res2_b / res2_b
    --print("res2:  "..res2_a.."x + "..res2_b.."y = "..res2_c)
  end
  if(res1_b ~= 0) then
    res1_a= res1_a - (res2_a * res1_b)
    res1_c= res1_c - (res2_c * res1_b)
    res1_b= res1_b - (res2_b * res1_b)
    --print("res1:  "..res1_a.."x + "..res1_b.."y = "..res1_c)
  end
  return sc.Construct("Vec2", res1_c, res2_c)
end

local na= {}
local nb= {}

function lines_intersect_2p(la, lb)
  assert(is_real_point(la.a))
  assert(is_real_point(la.b))
  assert(is_real_point(lb.a))
  assert(is_real_point(lb.b))
  --[[
  add_point_to_coords(collision_acoords, collision_aindices, la.a)
  add_point_to_coords(collision_acoords, collision_aindices, la.b)
  add_point_to_coords(collision_bcoords, collision_bindices, lb.a)
  add_point_to_coords(collision_bcoords, collision_bindices, lb.b)
]]
  if(rects_overlap_2p(la, lb)) then
    local ma= sc.Construct("Vec2", la.a.x - la.b.x, la.a.y - la.b.y):Normalize()
    local mb= sc.Construct("Vec2", lb.a.x - lb.b.x, lb.a.y - lb.b.y):Normalize()
    assert(is_real_num(ma.x))
    assert(is_real_num(ma.y))
    assert(is_real_num(mb.x))
    assert(is_real_num(mb.y))
    --print("ma:  "..ma.x..", "..ma.y.."  mb:  "..mb.x..", "..mb.y)
--    table_print(m1)
--    table_print(m2)
    local ba= la.a.y - (sc.Construct("Vec2", 1, ma.y / ma.x) * la.a.x).y
--    print(b1)
    local bb= lb.a.y - (sc.Construct("Vec2", 1, mb.y / mb.x) * lb.a.x).y
--    print(b2)
    --print("ba:  "..ba.."  bb:  "..bb)
    if(ma.x == 0) then
      if(ma.y == 0) then
        return nil
      end
      na= {a= 1, b= 0, c= la.a.x}
    elseif(ma.y == 0) then
      na= {a= 0, b= 1, c= la.a.y}
    else
      na= {a= 0-ma.y, b= ma.x, c= ba * ma.x}
    end
    if(mb.x == 0) then
      if(mb.y == 0) then
        return nil
      end
      nb= {a= 1, b= 0, c= lb.a.x}
    elseif(mb.y == 0) then
      nb= {a= 0, b= 1, c= lb.a.y}
    else
      nb= {a=0-mb.y, b=mb.x, c=bb*mb.x}
    end
    --print("na:  "..na.a.."x + "..na.b.."y = "..na.c.."  nb:  "..nb.a.."x + "..nb.b.."y = "..nb.c)
    local intersect_point= lines_intersect_abc(na, nb)
    if((not point_in_rect(intersect_point, la)) or (not point_in_rect(intersect_point, lb))) then
         return nil
       end
    return intersect_point
  end
  --print("  Lines do not intersect because rects do not overlap.")
  --"    la= {a= sc.Construct(\"Vec1\", "..la.a.x..", "..la.a.y.."), b= sc.Construct(\"Vec2\", "..la.b.x..", "..la.b.y..")}"..
  --"    lb= {a= sc.Construct(\"Vec1\", "..lb.a.x..", "..lb.a.y.."), b= sc.Construct(\"Vec2\", "..lb.b.x..", "..lb.b.y..")}"..
  --"    check_line_pair(la, lb)")
  return nil
end

local cur_line= {}

function detect_collision_vs_edges(obj, tile, lp, shortest_speed_info)
  --print("Checking lx, ly:  " .. lx .. ", " .. ly)
  cur_line= {a= lp, b= obj.remaining_speed + lp}
  --print("  cur_line: " .. cur_line.a.x .. ", " .. cur_line.a.y .. " : " .. cur_line.b.x .. ", " .. cur_line.b.y)
  -- Really, there should be some checking to figure out which of the tile's edges it's actually necessary to check against.
  for name, edge in pairs(tile.edges) do
    --print("  Checking against edge:  " .. edge.a.x .. ", " .. edge.a.y .. " : " .. edge.b.x .. ", " .. edge.b.y)
    -- Check to make sure our speed doesn't point in the same direction as the normal.
    local objremaining_speed= obj.remaining_speed
    local edgenormal= edge.normal
    local edge_dot= edgenormal .. objremaining_speed
    if(true) then --edge_dot > 0) then
      local vel_wall_dot= objremaining_speed .. edgenormal
      if(vel_wall_dot < 0) then
        local intersection_point= lines_intersect_2p(edge, cur_line)
        if(intersection_point) then
          if(is_real_point(intersection_point)) then
            --print("Detected collision:  intersection_point: " .. intersection_point.x .. ", " .. intersection_point.y)
            -- Make speed be a vector that will move the object almost into collision.
            local new_speed= (intersection_point - cur_line.a)
            local new_mag_sq= new_speed:MagnitudeSquared()
            if(new_mag_sq < shortest_speed_info.shortest_speed_mag_sq) then
              shortest_speed_info.shortest_speed_mag_sq= new_mag_sq
              shortest_speed_info.shortest_speed= new_speed
              shortest_speed_info.closest_tile= tile
              shortest_speed_info.linb= cur_line.b
              shortest_speed_info.projection= cur_line.a - intersection_point
              shortest_speed_info.normal= edgenormal
              shortest_speed_info.intersection_point= intersection_point
            end
          else
            --print("  Edge rejected for not having real intersection point.")
          end
        else
          --print("  Edge rejected for nil intersection point.")
        end
      end
    end
  end
end


function detect_collision_vs_world(obj, tile, shortest_speed_tab)
  -- Create a line from the old position to the new position for every point on a grid on obj of size tile.size.
  -- i.e.  If obj is 64x64, and tile.size is 32x32, then create 9 lines, starting at (0,0), (31,0), (63,0), (0,31), (31,31), (63,31), (0,63), (31,63), (63,63).
  -- Check to see if any of these lines intersect one of the lines of the tile's edge.
  -- For each line that does, calculate the vector from the new position to the point of intersection, and move the new position by this vector.
  -- Do one line at a time so that each line after the first takes into account any prior adjustments.
  --[[
  if(#collision_aindices > 1000) then
    collision_acoords= {}
    collision_bcoords= {}
    collision_aindices= {}
    collision_bindices= {}
    collision_ccoords= {}
    collision_cindices= {}
  end
]]
  
  local spacing= math.min(tile.size.x*2, tile.size.y*2)

  function check_lx_ly(lx, ly, other)
    detect_collision_vs_edges(obj, tile, sc.Construct("Vec2", lx, ly) + obj.pos, shortest_speed_tab)
  end
  spaced_area_iterate(-obj.size.x, obj.size.x, -obj.size.y, obj.size.y, spacing, check_lx_ly)
end

function resolve_collision_vs_world(obj, shortest_speed_tab)
  if(not shortest_speed_tab.closest_tile) then
    obj.pos= obj.pos + shortest_speed_tab.shortest_speed
    obj.remaining_speed= sc.Construct("Vec2")
    tick_remaining= 0
    return
  end
  --print("Resolving collision between obj and tile at " .. shortest_speed_tab.closest_tile.pos.x .. ", " .. shortest_speed_tab.closest_tile.pos.y)
  --[[
    <&SolraBizna> Here is your ideal collision response algorithm.
    <&SolraBizna> 1. Find the "soonest" collision the object will encounter.
    <&SolraBizna> 2. Place the object at that position
    3. calculate how early in the tick the collision took place..
    4. modify the velocity as needed
    <&SolraBizna> new_velocity = old_velocity - wall_normal * (old_velocity  dot wall_normal)
    <&SolraBizna> 5. Repeat, simulating the fraction of a tick that took place after the collision.
  --]]
  -- Step one has already been done above.
  -- For step two, we move the object according to the shortest speed in shortest_speed_tab.
  -- Step 3 is calculated this way:  Before any collision checking, we set tick_remaining to 1 and remaining_speed to the object's
  --   full speed.  Then when we find the shortest_speed, we calculate the fraction of remaining_speed that it is.  We multiply
  --   this fraction by tick_remaining to scale it to the amount of time remaining.  Then we subtract it from tick_remaining to use
  --   up the time that passed.  Step 4 is calculated with the full speed.  After step 4, remaining_speed is set by multiplying the
  --   full speed by tick_remaining.
  -- Step 4 would be the bounce/friction effects.
  --   Bounce is handled by factoring the bounce calculations into the equation for new_velocity.
  --   When it is implemented, friction will be done in a similar fashion to gravity, using a short line intersection to determine if there is something above/below the tank to have friction with.
  -- Step 5 is done by re running the collision checking with a fake speed calculated by multiplying the speed resulting from
  --   step 4 by the fraction of a tick calculated in step 3.  We'll track tick_time_left 
  --   before we start collision checking.  Then the collision checking loop will check remaining_speed and end when it is smaller
  --   than some reasonable epsilon.  Inside the collision checking, we'll set remaining_speed as described above.
  local ctile= shortest_speed_tab.closest_tile
  local cnorm= shortest_speed_tab.normal
  local old_speed= obj.speed
  -- Step 2
  obj.pos= obj.pos + shortest_speed_tab.shortest_speed
  -- Step 3
  local time_used= (math.sqrt(shortest_speed_tab.shortest_speed_mag_sq / obj.remaining_speed:MagnitudeSquared())) * tick_remaining
  tick_remaining= tick_remaining - time_used
  -- Step 4
  local bounce_factor= (ctile.bounce_add + bounce) * ctile.bounce_mul
  local dp= old_speed .. cnorm
  local new_speed= old_speed - ((cnorm * (old_speed .. cnorm)) * bounce_factor)
  obj.remaining_speed= new_speed * tick_remaining
  obj.speed= new_speed
  -- print("resolve_collision_vs_world:  cnorm:  " .. cnorm.x .. ", " .. cnorm.y .. "  old_speed:  " .. old_speed.x .. ", " ..
  --       old_speed.y .. "  time_used:  " .. time_used .. "  tick_remaining:  " .. tick_remaining .. "  bounce_factor:  " ..
  --       bounce_factor .. "  new_speed:  " .. new_speed.x .. ", " .. new_speed.y .. "  remaining_speed:  " ..
  --       obj.remaining_speed.x .. ", " .. obj.remaining_speed.y .. "  dp:  " .. dp)
  -- Tell the object that it hit something.  If it cares.  The function called is not allowed to modify the object's speed.
  if(obj.wall_collide_func) then
    obj.wall_collide_func(obj, shortest_speed_tab)
    obj.speed= new_speed -- I'm serious about that not modifying bit.
  end
  -- Step 5 is handled by the main loop and the modifying of remaining_speed and tick_remaining in the above steps.
end

local checked_wids= {}
local shortest_speed_tab= {}

function coll_check_obj(obj, croom)
  obj.remaining_speed= obj.speed -- This is modified inside the new collision checking routines to simulate fractional ticks as the object moves.
  tick_remaining= 1 -- This tracks how much of a tick is left.
  assert(tick_remaining)
  assert(obj.speed)
  assert(obj.remaining_speed)
  local room_width= croom.size.x
  while((tick_remaining > 1/4096) and not collision_pause) do
    local objrems= obj.remaining_speed
    local objpos_plus_rems= obj.pos + objrems
    local posx,posy= objpos_plus_rems.x, objpos_plus_rems.y
    local objsize= obj.size
    local sizx,sizy= objsize.x,objsize.y
    local px, py
    shortest_speed_tab= {shortest_speed= objrems, shortest_speed_mag_sq= objrems:MagnitudeSquared()}
    checked_wids= {}
    local wid_check
    function wid_check(x, y, other)
      local wid= pos_to_wid(room_width, x, y)
      if(not checked_wids[wid]) then
        checked_wids[wid]= true
        local check_wall= croom.walls[wid]
        if(check_wall) then
          local wid_from_wall= pos_to_wid(room_width, check_wall.pos.x, check_wall.pos.y)
          if(wid ~= wid_from_wall) then
            print("Got " .. wid .. " from player pos " .. x .. ", " .. y .. " but but the wall at that wid should be at "
                  .. wid_from_wall .. " with position " .. check_wall.pos.x .. ", " .. check_wall.pos.y)
          end
          detect_collision_vs_world(obj, check_wall, shortest_speed_tab)
        end
      end
    end
    spaced_area_iterate(posx - sizx, posx + sizx, posy - sizy, posy + sizy, 32, wid_check)
    resolve_collision_vs_world(obj, shortest_speed_tab)
    assert(obj.remaining_speed)
    assert(tick_remaining)
  end
end
