--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

function commodities_init()
  make_eq_fast= create_maker(5, {}, {"Equipment"}, 30)
  make_eq_med= create_maker(3, {}, {"Equipment"}, 30)
  make_eq_slow= create_maker(1, {}, {"Equipment"}, 30)
  make_food_fast= create_maker(10, {}, {"Food"}, 30)
  make_food_med= create_maker(6, {}, {"Food"}, 30)
  make_food_slow= create_maker(3, {}, {"Food"}, 30)
  make_med_fast= create_maker(5, {}, {"Medical"}, 60)
  make_med_med= create_maker(3, {}, {"Medical"}, 60)
  make_med_slow= create_maker(1, {}, {"Medical"}, 60)
  eat_eq_fast= create_maker(5, {{amount= 1, name= "Equipment"}}, {}, 30)
  eat_eq_med= create_maker(3, {{amount= 1, name= "Equipment"}}, {}, 30)
  eat_eq_slow= create_maker(1, {{amount= 1, name= "Equipment"}}, {}, 30)
  eat_food_fast= create_maker(10, {{amount= 1, name= "Food"}}, {}, 30)
  eat_food_med= create_maker(6, {{amount= 1, name= "Food"}}, {}, 30)
  eat_food_slow= create_maker(3, {{amount= 1, name= "Food"}}, {}, 30)
  eat_med_fast= create_maker(5, {{amount= 1, name= "Medical"}}, {}, 60)
  eat_med_med= create_maker(3, {{amount= 1, name= "Medical"}}, {}, 60)
  eat_med_slow= create_maker(1, {{amount= 1, name= "Medical"}}, {}, 60)

  comm_eq_low= create_comm("Equipment", 50, 60, 0.03)
  comm_eq_med= create_comm("Equipment", 25, 90, 0.06)
  comm_eq_high= create_comm("Equipment", 10, 120, 0.12)
  comm_food_low= create_comm("Food", 100, 10, 0.03)
  comm_food_med= create_comm("Food", 50, 13, 0.04)
  comm_food_high= create_comm("Food", 10, 16, 0.06)
  comm_med_low= create_comm("Medical", 40, 100, 0.06)
  comm_med_med= create_comm("Medical", 20, 200, 0.09)
  comm_med_high= create_comm("Medical", 5, 400, 0.12)
end

function create_comm(name, amount, price, volatility)
  return {name= name, amount= amount, amount_last_tick= amount, price= price, float_price= price, volatility= volatility}
end

function create_maker(cap, reqs, prods, freq)
  return {capacity= cap, required_comms= reqs, prod_comms= prods, next_make_time= 0, time_between_makes= freq}
end

function run_comm_maker(maker, planet)
  local prod_amount= maker.capacity
  for id, req in pairs(maker.required_comms) do
    local possible_amount= 0
    if planet.commodities[req.name] then
      possible_amount= planet.commodities[req.name].amount / req.amount
    end
    prod_amount= math.min(prod_amount, possible_amount)
  end
  if(prod_amount > 0) then
    for name, req in pairs(maker.required_comms) do
      planet.commodities[req.name].amount= planet.commodities[req.name].amount - (prod_amount * req.amount)
    end
    for id, name in pairs(maker.prod_comms) do
      if planet.commodities[name] then
        planet.commodities[name].amount= planet.commodities[name].amount + prod_amount
      else
        planet.commodities[name]= create_comm(name, prod_amount)
      end
    end
  end
  maker.next_make_time= gt_rticks + maker.time_between_makes
end

function comm_set_price(comm, planet)
  -- Eventually, this should take into account how quickly the commodity can be replaced if bought.
  local prev_price= comm.price
  -- If the amount this tick is greater than the amount last tick, we should have a lower price.
  -- If the amount this tick is less than the amount last tick, we should have a higher price.
  -- Both of these effects should be proportional to the size of the change.
  -- 0.3 should be replaced with something calculated from how much it costs to produce the commodity and/or some inherent value.
  comm.float_price= comm.float_price - ((comm.amount - comm.amount_last_tick) * comm.volatility)
  comm.price= math.ceil(comm.float_price)
end
