--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

require "planets"

local inf= 1/0

function common_init()
  planets_init()
end

function fbuy(buyer, seller, comm, amount, price)
  amount= math.min(amount, seller.commodities[comm].amount)
  buyer.credits= buyer.credits - (amount * price)
  buyer.ship.cargo_space= buyer.ship.cargo_space - amount
  if(buyer.ship.cargos[comm]) then
    buyer.ship.cargos[comm].amount= buyer.ship.cargos[comm].amount + amount
  else
    buyer.ship.cargos[comm]= {}
    buyer.ship.cargos[comm].amount= amount
  end
  seller.commodities[comm].amount= seller.commodities[comm].amount - amount
end

function fsell(seller, buyer, comm, amount, price)
  seller.credits= seller.credits + (amount * price)
  seller.ship.cargo_space= seller.ship.cargo_space + amount
  seller.ship.cargos[comm].amount= seller.ship.cargos[comm].amount - amount
  if(seller.ship.cargos[comm].amount == 0) then
    seller.ship.cargos[comm]= nil
  end
  buyer.commodities[comm].amount= buyer.commodities[comm].amount + amount
end

function table_copy(foo)
  local ret= {}
  for name, value in pairs(foo) do
    if(type(value) == "table") then
      ret[name]= table_copy(value) --[[ God bless recursion!  And if you make a
        self-referencing table, you deserve what you get.
        "Have I been here before?"  "I can't tell, it all looks the same to me." ]]--
    else
      ret[name]= value
    end
  end
  return ret
end

function table_print(foo, indent)
  indent= indent and indent or "" -- Looks strange, but sets indent to empty string if it's nil
  for name, value in pairs(foo) do
    if(type(value) == "table") then
      print(indent .. name .. ":")
      table_print(value, indent .. "  ") --[[ God bless recursion!  And if you make a
        self-referencing table, you deserve what you get.
        "Have I been here before?"  "I can't tell, it all looks the same to me." ]]--
    elseif(type(value) == "function") then
      print(indent .. "function: " .. name)
    elseif(type(value) == "userdata") then
      print(indent .. "userdata: " .. name)
    elseif(type(value) == "boolean") then
      print(indent .. "boolean: " .. name)
    else
      print(indent .. name .. ":  " .. value)
    end
  end
  print(indent .. "table end")
end

function global_print()
  for name, value in pairs(_G) do
    if(type(value) == "table") then
      if(name == "_G") then
        -- do nothing
      else
        print("table: " .. name)
        table_print(value)
      end
    elseif(type(value) == "function") then
      print("function: " .. name)
    elseif(type(value) == "userdata") then
      print("userdata: " .. name)
    elseif(type(value) == "boolean") then
      print("boolean: " .. name)
    else
      print(name .. ":  " .. value)
    end
  end
end

function rect_print(r, pre)
--  print(type(r.x) .. type(r.y) .. type(r.w) .. type(r.h))
  if(pre) then
    print(pre .. r.x .. ", " .. r.y .. ", " .. r.w .. ", " .. r.h)
  else
    print(r.x .. ", " .. r.y .. ", " .. r.w .. ", " .. r.h)
  end
end

function xor(a, b) -- It's simply astounding that I have to write this function,
  -- instead of it being part of the language like and, or and not.
  return (a and not b) or (b and not a)
end

function table.cat(a, b) -- this function changes a!
  for n=1,table.getn(b) do
    table.insert(a, b[n])
  end
  return a
end

function scale_vector(v, s)
  return {x= v.x * s, y= v.y * s}
end

function unit_vector(v)
  local mag= math.sqrt((v.x * v.x) + (v.y * v.y))
  return {x= v.x / mag, y= v.y / mag}
end

function vector_mag(v)
  return math.sqrt((v.x * v.x) + (v.y * v.y))
end

function add_two_points(a, b)
  return {x= a.x + b.x, y= a.y + b.y}
end

function is_real_num(n)
  return not (n == inf or n == -inf or n ~= n)
end

function is_real_point(p)
  return not ((p.x == inf or p.x == -inf) or (p.y == inf or p.y == -inf) or (p.x ~= p.x or p.y ~= p.y))
end

function minmax(a, b)
  if(a > b) then
    return b, a
  end
  return a, b
end

function minmax4(a, b, c, d)
  local i, a= minmax(a, b)
  local n, x= minmax(c, d)
  return math.min(i, n), math.max(a, x)
end

function min_abs3(a, b, c)
  local absa= math.abs(a)
  local absb= math.abs(b)
  local absc= math.abs(c)
  if absa < absb then
    if absa < absc then
      return a
    else
      return c
    end
  else
    if absb < absc then
      return b
    else
      return c
    end
  end
end

function pos_to_id(xoff, yoff, width, spacing, x, y)
  return math.floor((x-xoff)/spacing) + (math.floor((y-yoff)/spacing)*width)
end

function pos_to_wid(room_width, x, y)
  return math.floor(x/32) + (math.floor(y/32) * room_width)
end

function spaced_area_iterate(sx, ex, sy, ey, spacing, callback, callback_args)
  local cx, cy
  for cx= sx, ex, spacing do
    for cy= sy, ey, spacing do
      callback(cx, cy, callback_args)
    end
    if(math.mod(ey - sy, spacing) ~= 0) then
      cy= ey
      callback(cx, cy, callback_args)
    end
  end
  if(math.mod(ex - sx, spacing) ~= 0) then
    cx= ex
    for cy= sy, ey, spacing do
      callback(cx, cy, callback_args)
    end
    if(math.mod(ey - sy, spacing) ~= 0) then
      cy= ey
      callback(cx, cy, callback_args)
    end
  end
end

function no_edge_spaced_area_iterate(sx, ex, sy, ey, spacing, callback, callback_args)
  local cx, cy
  for cx= sx, ex, spacing do
    for cy= sy, ey, spacing do
      callback(cx, cy, callback_args)
    end
  end
end

function spaced_line_iterate(pos, speed, spacing, callback, callback_args)
  local space_vec= speed:Normalize() * spacing
  local speed_mag= speed:Magnitude()
  local iterations= math.ceil(speed_mag / spacing) + 1
  local cp= pos
  for ci= 1, iterations, 1 do
    callback(cp, callback_args)
    cp= cp + space_vec
    -- We don't need to make sure we check the end of the line because we set iterations to ensure we overshoot and overshooting is okay where this function is used.
  end
end

function add_point_to_coords(coords, indices, point)
  coords[#coords+1]= point.x
  coords[#coords+1]= point.y
  indices[#indices+1]= (#coords-1)/2
end

function add_point_only_to_coords(coords, point)
  coords[#coords+1]= point.x
  coords[#coords+1]= point.y
end

function set_cholera(drawable, cholera)
  if(cholera.a) then
    drawable:SetPrimitiveColor(cholera.r, cholera.g, cholera.b, cholera.a)
  else
    drawable:SetPrimitiveColor(cholera.r, cholera.g, cholera.b)
  end
end

