--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

-- This is for generating maps with a room connection layout similar to Kyzentun's sigil.
-- It is based on the observation that the sigil is topologically equivalent to a set of rows offset from each other and connected.
-- Each row is built connected to the ones above and below.
-- Then the rooms are filled with the wall data.

gamelicense = "Compatible"
require "subcritical"

require "common"
require "map"

local row_size= 3
local num_rows= 3

function gen_conned_rooms()
  local ret= {}
  local prev_rows
  for prev_rows= 0, num_rows-1, 1 do
    local cur_row_start= prev_rows * row_size
    local prev_row_start= (prev_rows - 1) * row_size
    if(prev_rows == 0) then
      prev_row_start= (num_rows - 1) * row_size
    end
    local next_row_start= (prev_rows + 1) * row_size
    if(prev_rows == num_rows - 1) then
      next_row_start= 0
    end
    local n
    for n=0, row_size-1, 1 do
      local ls, ts, rs, bs
      local lr, tr, rr, br
      if(n == 0) then
        ls= prev_row_start
        lr= 3
        bs= next_row_start
        br= 1
      else
        ls= cur_row_start + n - 1
        lr= 0
        bs= next_row_start + n - 1
        br= 0
      end
      if(n == row_size-1) then
        rs= next_row_start + n
        rr= 3
        ts= prev_row_start + n
        tr= 1
      else
        rs= cur_row_start + n + 1
        rr= 0
        ts= prev_row_start + n + 1
        tr= 0
      end
      ret[cur_row_start+n]= {conns= {left= {side= ls, rot= lr}, top= {side= ts, rot= tr}, right= {side= rs, rot= rr}, bottom= {side= bs, rot= br}},
                             places= {}}
    end
  end
  return ret
end

function gen_subroom(dest_room_array, x, y, xmin_open, ymin_open, xmax_open, ymax_open)
  -- Pick a width for each opening and center the opening on its side.
  local xminoh= math.random(1, 16)
  local yminoh= math.random(1, 16)
  local xmaxoh= math.random(1, 16)
  local ymaxoh= math.random(1, 16)
  if not xmin_open then
    xminoh= 0
  end
  if not ymin_open then
    yminoh= 0
  end
  if not xmax_open then
    xmaxoh= 0
  end
  if not ymax_open then
    ymaxoh= 0
  end
  -- Draw rectangles of closed space in the corners.
  local xmin_margin_size= (16 - xminoh) / 2
  local ymin_margin_size= (16 - yminoh) / 2
  local xmax_margin_size= (16 - xmaxoh) / 2
  local ymax_margin_size= (16 - ymaxoh) / 2

  local ulrxs= math.floor(ymin_margin_size)
  local ulrys= math.ceil(xmin_margin_size)
  local urrxs= math.ceil(ymin_margin_size)
  local urrys= math.ceil(xmax_margin_size)
  local lrrxs= math.ceil(ymax_margin_size)
  local lrrys= math.floor(xmax_margin_size)
  local llrxs= math.floor(ymax_margin_size)
  local llrys= math.floor(xmin_margin_size)
  function write_solid_rect(array, xs, ys, xe, ye)
    for x= xs, xe, 1 do
      for y= ys, ye, 1 do
        array[x][y]= 1
      end
    end
  end
  write_solid_rect(dest_room_array, x, y, x + ulrxs-1, y + ulrys-1)
  write_solid_rect(dest_room_array, x + ulrxs + yminoh, y, x + 15, y + urrys-1)
  write_solid_rect(dest_room_array, x + llrxs + ymaxoh, y + urrys + xmaxoh, x + 15, y + 15)
  write_solid_rect(dest_room_array, x, y + ulrys + xminoh, x + llrxs-1, y + 15)
  if xminoh >= 5 then
    local ledge_x= math.random(1, 7)
    local ledge_y= math.random(6, 10)
    write_solid_rect(dest_room_array, x+ledge_x, y+ledge_y, x+ledge_x, y+ledge_y)
  end
  if yminoh >= 5 then
    local ledge_x= math.random(6, 10)
    local ledge_y= math.random(1, 7)
    write_solid_rect(dest_room_array, x+ledge_x, y+ledge_y, x+ledge_x, y+ledge_y)
  end
  if xmaxoh >= 5 then
    local ledge_x= math.random(8, 14)
    local ledge_y= math.random(6, 10)
    write_solid_rect(dest_room_array, x+ledge_x, y+ledge_y, x+ledge_x, y+ledge_y)
  end
  if ymaxoh >= 5 then
    local ledge_x= math.random(6, 10)
    local ledge_y= math.random(8, 14)
    write_solid_rect(dest_room_array, x+ledge_x, y+ledge_y, x+ledge_x, y+ledge_y)
  end
end

function gen_room(dest_room)
-- Generate the center 4 as open on all sides.
-- Expand outward from them in the open directions.
-- When generating a subroom, make it open on the side the algorithm came from and give a high chance of the
-- opposite side being open.

  local genned_subrooms= {}
  -- Info on each generated subroom is stored in a table.  The index of the table is the concatenation of the
  -- string forms of its coordinates.
  genned_subrooms[tostring(3)..tostring(3)]= {x= 3, y= 3, xmino= true, ymino= true, xmaxo= true, ymaxo= true}
  genned_subrooms[tostring(3)..tostring(4)]= {x= 3, y= 4, xmino= true, ymino= true, xmaxo= true, ymaxo= true}
  genned_subrooms[tostring(4)..tostring(3)]= {x= 4, y= 3, xmino= true, ymino= true, xmaxo= true, ymaxo= true}
  genned_subrooms[tostring(4)..tostring(4)]= {x= 4, y= 4, xmino= true, ymino= true, xmaxo= true, ymaxo= true}

  local gen_next_list= {{x= 3, y= 2}, {x= 4, y= 2},
            {x= 2, y= 3},                         {x= 5, y= 3},
            {x= 2, y= 4},                         {x= 5, y= 4},
                        {x= 3, y= 5}, {x= 4, y= 5},
   }
  while #gen_next_list > 0 do
    local genx= gen_next_list[1].x
    local geny= gen_next_list[1].y
    table.remove(gen_next_list, 1)
    if not genned_subrooms[tostring(genx)..tostring(geny)] then
      local xmino_chance= 70
      local ymino_chance= 70
      local xmaxo_chance= 70
      local ymaxo_chance= 70
      if genned_subrooms[tostring(genx-1)..tostring(geny)] then
        if genned_subrooms[tostring(genx-1)..tostring(geny)].xmaxo then
          xmino_chance= xmino_chance + 20
        else
          xmino_chance= xmino_chance - 20
        end
      elseif genx > 0 then
        gen_next_list[#gen_next_list+1]= {x= genx-1, y= geny}
      end
      if genned_subrooms[tostring(genx)..tostring(geny-1)] then
        if genned_subrooms[tostring(genx)..tostring(geny-1)].ymaxo then
          ymino_chance= ymino_chance + 20
        else
          ymino_chance= ymino_chance - 20
        end
      elseif geny > 0 then
        gen_next_list[#gen_next_list+1]= {x= genx, y= geny-1}
      end
      if genned_subrooms[tostring(genx+1)..tostring(geny)] then
        if genned_subrooms[tostring(genx+1)..tostring(geny)].xmino then
          xmaxo_chance= xmaxo_chance + 20
        else
          xmaxo_chance= xmaxo_chance - 20
        end
      elseif genx < 7 then
        gen_next_list[#gen_next_list+1]= {x= genx+1, y= geny}
      end
      if genned_subrooms[tostring(genx)..tostring(geny+1)] then
        if genned_subrooms[tostring(genx)..tostring(geny+1)].ymino then
          ymaxo_chance= ymaxo_chance + 20
        else
          ymaxo_chance= ymaxo_chance - 20
        end
      elseif geny < 7 then
        gen_next_list[#gen_next_list+1]= {x= genx, y= geny+1}
      end
      local xmino= (math.random(1, 100) <= xmino_chance)
      local ymino= (math.random(1, 100) <= ymino_chance)
      local xmaxo= (math.random(1, 100) <= xmaxo_chance)
      local ymaxo= (math.random(1, 100) <= ymaxo_chance)
      genned_subrooms[tostring(genx)..tostring(geny)]= {x= genx, y= geny, xmino= xmino, ymino= ymino, xmaxo= xmaxo, ymaxo= ymaxo}
    end
  end
  -- Fill the room with open space.
  local xs= 128
  local ys= 128
  dest_room.array= {size= {x= xs, y= ys}}
  for y= 1, ys, 1 do
    dest_room.array[y]= {}
    for x= 1, xs, 1 do
      dest_room.array[y][x]= 0
    end
  end
  -- Generate each of the subrooms.
  for k, v in pairs(genned_subrooms) do
    gen_subroom(dest_room.array, (v.x*16)+1, (v.y*16)+1, v.xmino, v.ymino, v.xmaxo, v.ymaxo)
  end
  -- Stick the subroom data in the room data structure so we can use it for the minimap later.
  dest_room.subrooms= genned_subrooms
end

function fill_rooms(rooms)
  -- Eventually, this needs to randomly place planets in the rooms and make paths between them.
  for number, room in pairs(rooms) do
    gen_room(room)
  end
end

function place_planets(rooms)
  -- Eventually, this will randomly generate planets and scatter them throughout the rooms.
  rooms[0].places= {}
  rooms[1].places= {{name= "Klyztam", pos= {x= 1792, y= 1792}, size= {x= 32, y= 32}, gin= 0, rot= 0},
                    {name= "Glorsap", pos= {x= 1280, y= 1280}, size= {x= 32, y= 32}, gin= 1, rot= 0},
                    {name= "Fhqwhgads", pos= {x= 1280, y= 1792}, size= {x= 32, y= 32}, gin= 2, rot= 0},
                    {name= "Twad_Savd", pos= {x= 1280, y= 2304}, size= {x= 32, y= 32}, gin= 3, rot= 0},
                    {name= "Nowhere", pos= {x= 1792, y= 1280}, size= {x= 32, y= 32}, gin= 4, rot= 3},
                    {name= "AEON", pos= {x= 2304, y= 1792}, size= {x= 32, y= 32}, gin= 5, rot= 3}}
  rooms[2].places= {}
  rooms[3].places= {}
  rooms[4].places= {}
  rooms[5].places= {}
  rooms[6].places= {}
  rooms[7].places= {}
  rooms[8].places= {}
end

function area_copy(src, src_sx, src_sy, src_ex, src_ey, dest)
  local dest_x= 1
  local dest_y= 1
  for src_x= src_sx, src_ex, 1 do
    dest[dest_x]= {}
    for src_y= src_sy, src_ey, 1 do
      dest[dest_x][dest_y]= src[src_x][src_y]
      dest_y= dest_y + 1
    end
    dest_x= dest_x + 1
  end
end

function build_minimap_from_array_room(room, dest, dw, dh)
  local minigraph= dest
  minigraph:SetPrimitiveColor(0, 0, 0)
  minigraph:DrawRect(0, 0, dw, dh)
  local room_walls= room.walls
  local wall_points= {}
  for x= 1, room.array.size.x, 1 do
    for y= 1, room.array.size.y, 1 do
      if room.array[y][x] ~= 0 then
        wall_points[#wall_points+1]= x-1
        wall_points[#wall_points+1]= y-1
      end
    end
  end
  minigraph:SetPrimitiveColor(1, 1, 1)
  minigraph:DrawPoints(SCUtil.CompileCoords(wall_points), 1)
end

function build_subroom_minimap(room, dest, dw, dh)
  dest:SetPrimitiveColor(1, 1, 1)
  dest:DrawRect(0, 0, dw, dh)
  dest:SetPrimitiveColor(0, 0, 0)
  for k, v in pairs(room.subrooms) do
    if v.xmino then
      dest:DrawRect((v.x*4), (v.y*4)+1, 2, 2)
    end
    if v.ymino then
      dest:DrawRect((v.x*4)+1, (v.y*4), 2, 2)
    end
    if v.xmaxo then
      dest:DrawRect((v.x*4)+2, (v.y*4)+1, 2, 2)
    end
    if v.ymaxo then
      dest:DrawRect((v.x*4)+1, (v.y*4)+2, 2, 2)
    end
  end
end

function output_rooms(rooms)
  pngdump = SubCritical.Construct("PNGDumper")
  for id, room in pairs(rooms) do
    local room_name= "rooms/"..id
    local minigraph= sc.Construct("Graphic", 128, 128)
    build_minimap_from_array_room(room, minigraph, 128, 128)
    local map_file, err= io.open(SCPath(room_name..".png"), "wb")
    if not map_file then
      print("Failed to open file to save room minimap:  "..err)
      return
    end
    pngdump:Dump(minigraph, map_file, map_file.write)
    map_file:close()
    local subminigraph= sc.Construct("Graphic", 32, 32)
    build_subroom_minimap(room, subminigraph, 32, 32)
    local sub_map_file, err= io.open(SCPath(room_name.."sub.png"), "wb")
    if not sub_map_file then
      print("Failed to open file to save subroom minimap:  "..err)
      return
    end
    pngdump:Dump(subminigraph, sub_map_file, sub_map_file.write)
    sub_map_file:close()
    local room_file, err= io.open(SCPath(room_name..".celduin"), "wb")
    if not room_file then
      print("Failed to open file to save room:  "..err)
      return
    end
    local celduin= SCUtil.Dehydrate(room)
    -- Try to rehydrate it to make sure we've created a valid celduin.
    local rehyd_room= SCUtil.Rehydrate(celduin)
    local checksum= SCUtil.QuickChecksum(celduin)
    room_file:write(checksum)
    room_file:write(celduin)
    room_file:close()
  end
end

math.randomseed(os.time())

local all_rooms= gen_conned_rooms()
print("Done generating connected rooms.")
fill_rooms(all_rooms)
print("Done filling rooms.")
place_planets(all_rooms)
print("Done placing bases.")
output_rooms(all_rooms)
print("Done saving rooms.")
