--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

require "scui"
require "uranium"

local list_widget= {}

function list_widget:RenderSelf(window)
  assert(self.w and self.h, "list with no width/height?")
  print("list_widget:RenderSelf")

  if(self.outline_color) then
    set_cholera(screen, self.outline_color)
  else
    screen:SetPrimitiveColor(1, 1, 1)
  end
  screen:DrawBox(self._x, self._y, self.w, self.h)
  if(self.inside_color) then
    set_cholera(screen, self.inside_color)
  else
    screen:SetPrimitiveColor(0, 0, 0)
  end
  screen:DrawRect(self._x+1, self._y+1, self.w-2, self.h-2)
  if(self.indicator) then
    if(self.indicator.color) then
      set_cholera(screen, self.indicator.color)
    else
      screen:SetPrimitiveColor(0, .5, 0)
    end
    screen:DrawRect(self._x+2, self.indicator._y, self.w-4, self.indicator.h)
  end
  local x,y = self._x+2,self._y+2
  assert(self.columns, "list object with no text?!")
  if(not self.friskets_set) then
    self:SetupText()
  end
  local skip = self.skip
  screen:SetPrimitiveColor(1, 1, 1)
  for i= 1, #self.columns, 1 do
    local reny= y
    local remh = self.h
    local fs= self.columns[i].friskets
    for f= 1, #fs, 1 do
      if(remh <= 0) then return end
      if(fs[f]) then
        screen:BlitFrisket(fs[f], 0, 0, self.columns[i].width, math.ceil(remh), x + self.columns[i].pos, math.floor(reny))
      end
      reny = reny + skip
      remh = remh - skip
    end
  end
end

function list_widget:SetupText()
  local biggest_w = 0
  if(not self.font) then self.font = uranium.font end
  local font= self.font
  local skip = font:GetLineHeight()
  self.skip = skip
  self.friskets_set= true
  for i= 1, #self.columns, 1 do
    local n = 1
    local start = 1
    local y = 0
    local text = self.columns[i].text
    self.columns[i].friskets = {}
    repeat
      local stop,nustart = font:BreakLine(text, self.w, start)
      if(stop) then
        if(stop < start) then
          self.columns[i].friskets[n] = false
        else
          self.columns[i].friskets[n] = font:RenderText(text, start, stop, self.columns[i].width, 0, y)
          local w,h = self.columns[i].friskets[n]:GetSize()
          w= w + self.columns[i].pos
          if(w > biggest_w) then biggest_w = w end
        end
        n = n + 1
        y = y + skip
        start = nustart
      end
    until not start
  end
  if(not self.w) then self.w = biggest_w end
  if(not self.h) then self.h = math.ceil(y) end
end

function list_widget:action()
  return self.tag
end

function list_widget:OnHotKey(key, mod)
  if(not key) then print ("list_widget:OnHotKey key is nil") end
  if(not mod) then print ("list_widget:OnHotKey mod is nil") end
  --print("list_widget:OnHotKey:  key = " .. key)
  if(self.indicator) then
    if(key == self.upkey) then
      if(self.indicator.sel > self.min_sel) then
        self.indicator.sel = self.indicator.sel - 1
        self.indicator._y = self.indicator._y - self.indicator.h
        scui.MarkDirty(self)
      end
    elseif(key == self.downkey) then
      if(self.indicator.sel < self.max_sel) then
        --print("list_widget:OnHotKey:  key = " .. key .. " == " .. self.downkey)
        self.indicator.sel = self.indicator.sel + 1
        self.indicator._y = self.indicator._y + self.indicator.h
        scui.MarkDirty(self)
      end
    end
  end
  return nil
end

function list_widget:OnMouseDown()
  if(not self.indicator) then return nil end
  local x,y= screen:GetMousePos()
  local choice= math.ceil((y - (self._y + 2)) / self.indicator.h)
  if(choice >= self.min_sel and choice <= self.max_sel) then
    self.indicator.sel= choice
    self.indicator._y= self._y + 2 + (self.indicator.h * (choice - 1))
    scui.MarkDirty(self)
  end
  return nil
end

function gp_start()
  cp= planets[player.cur_planet]
  plc= cp.commodities
  local num_commodities= 0
  index_to_comm_name= {}
  for name, value in pairs(plc) do
    num_commodities= num_commodities + 1
    index_to_comm_name[num_commodities] = name
  end
  local nt, pt, psat, pat= build_comm_text()
  local pcomm_text= ""
  for name, value in pairs(player.ship.cargos) do
    pcomm_text= pcomm_text .. name .. " | " .. value.amount .. "\n"
  end
  gp_widget= {
    x= 0, y= 0, w= screen_width, h= screen_height,
    type= uranium.box,
    filled= true,
    {x= 16, y= 32, type= uranium.text, text= player.cur_planet, font= font_lg},
    {x= 16, y= 64, type= uranium.text, text= cp.desc, font= font_sm},
    {x= 16, y= 96, type= uranium.text, text= "Commodities market:", font= font_med},
    {x= 16, y= 112, type= uranium.text, text= "Name | price | amount available | amount in hold", font= font_med},
    {x= 16, y= 128, w= 256, h= 128, type= list_widget,
     columns= {{text= nt, pos= 0, width= 96}, {text= pt, pos= 96, width= 48}, {text= pat, pos= 144, width= 48},
               {text= psat, pos= 192, width= 48}},
     font= font_med, upkey= "up", downkey= "down",
     hotkeys= {["up"]=true, ["down"]=true}, min_sel= 1, max_sel= num_commodities,
     indicator= {_y= 130, h= 14, sel= 1}},
    {x= 16, y= 256, w= 48, h= 16, type= uranium.checkbox, text= "/10", hotkeys= {["4"]= true}, checked= down_mod},
    {x= 64, y= 256, w= 48, h= 16, type= uranium.checkbox, text= "*10", hotkeys= {["1"]= true}, checked= up_mod1},
    {x= 112, y= 256, w= 48, h= 16, type= uranium.checkbox, text= "*100", hotkeys= {["2"]= true}, checked= up_mod2},
    {x= 160, y= 256, w= 48, h= 16, type= uranium.checkbox, text= "max", hotkeys= {["3"]= true}, checked= up_mod3},
    {x= 16, y= 288, w= 48, h= 20, type= uranium.button, text= "Buy", action= gp_buy, hotkeys= {["right"]= true}},
    {x= 64, y= 288, w= 48, h= 20, type= uranium.button, text= "Sell", action= gp_sell, hotkeys= {["left"]= true}},
    {x= 16, y= 320, w= 48, h= 20, type= uranium.button, text= "Refuel", action= gp_refuel, hotkeys= {["r"]= true}},
    {x= 64, y= 320, w= 48, h= 20, type= uranium.button, text= "Leave", action= gp_cleanup, hotkeys= {["escape"]= true}},
    {x= 616, y= 16, w= 112, h= 20, type= uranium.text, text= "Name:  " .. player.name, font= font_med},
    {x= 616, y= 32, w= 112, h= 20, type= uranium.text, text= "Credits:  " .. player.credits, font= font_med},
    {x= 616, y= 48, w= 112, h= 20, type= uranium.text, text= "Fuel:  " .. math.floor(player.ship.fuel), font= font_med},
    {x= 616, y= 64, w= 112, h= 20, type= uranium.text, text= "Cargo Space:  " .. player.ship.cargo_space, font= font_med},
    {x= 616, y= 80, w= 112, h= 128, type= uranium.box, filled= true, {x= 1, y= 1, w= 110, h= 126, type= uranium.text, text= pcomm_text, fnt= font_med}},
  }
  scui.CompileWidget(gp_widget)
  scui.RunModal(gp_widget, nil, uranium.cursor())
  -- RunModal returns when one of the functions attached to part of the dialog returns non-nil.
  gp_widget= nil
  index_to_comm_name= nil
  plc= nil
  cp= nil
  next_play_mode= "travel"
end

function gp_cleanup()
  return "Leaving."
end

function build_comm_text()
  local name_text= ""
  local prices_text= ""
  local player_amounts_text= ""
  local planet_amounts_text= ""
  for name, value in pairs(plc) do
    local pa= 0
    if player.ship.cargos[name] then
      pa= player.ship.cargos[name].amount
    end
    name_text= name_text .. name .. "\n"
    prices_text= prices_text .. value.price .. "\n"
    player_amounts_text= player_amounts_text .. pa .. "\n"
    planet_amounts_text= planet_amounts_text .. value.amount .. "\n"
  end
  return name_text, prices_text, player_amounts_text, planet_amounts_text
end  

function rebuild_comm_text()
  local nt, pt, psat, pat= build_comm_text()
  gp_widget[5].columns[1].text= nt
  gp_widget[5].columns[2].text= pt
  gp_widget[5].columns[3].text= pat
  gp_widget[5].columns[4].text= psat
  gp_widget[5].friskets_set= nil
  gp_widget[15].text= "Credits:  " .. player.credits
  gp_widget[15].friskets= nil
  gp_widget[17].text= "Cargo Space:  " .. player.ship.cargo_space
  gp_widget[17].friskets= nil
  local comm_text= ""
  for name, value in pairs(player.ship.cargos) do
    comm_text= comm_text .. name .. " | " .. value.amount .. "\n"
  end
  gp_widget[18][1].text= comm_text
  gp_widget[18][1].friskets= nil
  scui.CompileWidget(gp_widget)
end

function gp_set_mods()
  up_mod1= gp_widget[7].checked
  up_mod2= gp_widget[8].checked
  up_mod3= gp_widget[9].checked
  down_mod= gp_widget[6].checked
 end

function gp_buy()
  gp_set_mods()
  g_buy(index_to_comm_name[gp_widget[5].indicator.sel])
  rebuild_comm_text()
  return nil
end

function gp_sell()
  gp_set_mods()
  g_sell(index_to_comm_name[gp_widget[5].indicator.sel])
  rebuild_comm_text()
  return nil
end

function gp_refuel()
  gp_set_mods()
  g_refuel()
  gp_widget[15].text= "Credits:  " .. player.credits
  gp_widget[15].friskets= nil
  gp_widget[16].text= "Fuel:  " .. math.floor(player.ship.fuel)
  gp_widget[16].friskets= nil
  scui.CompileWidget(gp_widget)
end
