--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

require "image_manager"
require "map"
require "player"
require "gui_planet"
require "gui_travel"
require "map_screen"

font_lg = TTF.Liberation.Serif[24]
font_med = TTF.Liberation.Sans[12]
font_sm = TTF.Liberation.Serif[10]

local friskets,briskets,triskets = {},{},{}
local function clear_text_cache()
   friskets,briskets,triskets = {},{},{}
   repeat until collectgarbage("step")
end

function printf(x, y, format, ...)
   local text = format:format(...)
   local frisket = friskets[text]
   if(not frisket) then
      frisket = font_med:RenderText(text)
      friskets[text] = frisket
   end
   screen:BlitFrisket(frisket, x, y - font_med:GetLineHeight())
end

function bprintf(x, y, format, ...)
   local text = format:format(...)
   local frisket = briskets[text]
   if(not frisket) then
      frisket = font_lg:RenderText(text)
      briskets[text] = frisket
   end
   screen:BlitFrisket(frisket, x, y - font_lg:GetLineHeight())
end

function tprintf(x, y, format, ...)
   local text = format:format(...)
   local frisket = triskets[text]
   if(not frisket) then
      frisket = font_sm:RenderText(text)
      triskets[text] = frisket
   end
   screen:BlitFrisket(frisket, x, y - font_sm:GetLineHeight())
end

function bcprintf(y, format, ...)
   local text = format:format(...)
   local frisket = briskets[text]
   if(not frisket) then
      frisket = font_lg:RenderText(text)
      briskets[text] = frisket
   end
   screen:BlitFrisket(frisket, screen_width / 2 - frisket:GetSize() / 2, screen_height / 2 - y - font_lg:GetLineHeight())
end

function tcprintf(y, format, ...)
   local text = format:format(...)
   local frisket = triskets[text]
   if(not frisket) then
      frisket = font_sm:RenderText(text)
      triskets[text] = frisket
   end
   screen:BlitFrisket(frisket, screen_width / 2 - frisket:GetSize() / 2, screen_height / 2 - y - font_sm:GetLineHeight())
end

function create_point(x, y, z)
  local foo= {}
  foo.x= x
  foo.y= y
  foo.z= z
  return foo
end

function mod_set_num(max)
   local ret= 10
  if(down_mod == true) then
    ret= ret / 10
  end
  if(up_mod1 == true) then
    ret = ret * 10
  end
  if(up_mod2 == true) then
    ret = ret * 100
  end
  if(up_mod3 == true) then
    return max
  end
  return ret
end

function g_buy(comm)
  assert(comm, "Tried to g_buy with nil comm")
--  print("g_buy called: " .. comm)
  local ba= mod_set_num(player.ship.cargo_space)
--  print(ba)
  if(ba > player.ship.cargo_space) then
    ba= player.ship.cargo_space
  end
  if(ba * plc[comm].price > player.credits) then
    ba= math.floor(player.credits / plc[comm].price)
  end
--  print(ba)
  fbuy(player, cp, comm, ba, plc[comm].price)
--  print("foo!")
end

function g_sell(comm)
  assert(comm, "Tried to g_sell with nil comm")
--  print("g_sell called: " .. comm)
  if(player.ship.cargos[comm]) then
    local sa= mod_set_num(player.ship.cargos[comm].amount)
    --  print(sa)
    if(sa > player.ship.cargos[comm].amount) then
      sa= player.ship.cargos[comm].amount
    end
    --  print(sa)
    fsell(player, cp, comm, sa, plc[comm].price)
  end
end

function g_refuel(comm)
--  print("g_refuel called: " .. comm)
  if(player.credits < 100) then return end
  local ra= mod_set_num(player.ship.max_fuel - player.ship.fuel)
  if(ra > player.ship.max_fuel) then
    ra= player.ship.max_fuel
  end
  if(ra * 100 > player.credits) then
    ra= math.floor(player.credits / 100)
  end
  player.ship.fuel= player.ship.fuel + ra
  player.credits= player.credits - (ra * 100)
end

function comefrom(foo, bar)
--  print("Coming from " .. foo .. " line " .. bar)
  return
end

function g_land(p)
--  print("Docking at " .. p)
  gt_cleanup()
--  print("Travel cleaned up.")
  player.cur_planet= p
  gp_start()
--  print("Base started.")
end

function g_travel(extra)
--[[
--  print("g_travel called: " .. extra)
  if(cp.connections[extra]) then
    if(player.ship.fuel > 0) then
      player.cur_planet= extra
      player.ship.fuel= player.ship.fuel - 1
    else
      print("You have no fuel!")
    end
  else
    print("Your hyperdrive has overloaded!")
    tint.clearidleproc()
  end
  gr_pos.y= 1
  --]]
--  print("Cleaning up planet stuff.")
  gp_cleanup()
--  print("Starting travel stuff.")
  gt_start()
--  comefrom("start_room", 11)
--  print("And away we go.")
end

function g_init()
  ps= 12
  lh= ps+2
  map_init()
--  print("map inited")
  gui_player_init()
  print("You are " .. player.name)
  for name, value in pairs(planets) do
     value.image= png:Load(SCPath(value.img_file_name))
  end
  place_grc= png:Load(SCPath("base_pics.png"))
--  print("done")
end

function gs_keypress(key)
  if(key == "return") then
    return
  end
  str= str .. key -- Hm, this doesn't handle modifier keys, or special keys.  BUT IT WORKS!
  screen:SetPrimitiveColor(0, 0, 0)
  screen:DrawRect(str_pos.x, str_pos.y, screen_width, 20)
  screen:SetPrimitiveColor(1, 1, 1)
  tprintf(str_pos.x, str_pos.y, str)
end

function gs_keyrelease(key)
  if(key == "return") then
    str_not_done= nil
  end
end

function gui_get_str(prompt_pos, prompt_str)
--  global_print()
  screen:SetPrimitiveColor(1, 1, 1)
  printf(prompt_pos[1], prompt_pos[2], prompt_str)
  str_pos= {x= prompt_pos[1], y= prompt_pos[2] + lh, z= prompt_pos[3]}
  str= ""
  str_not_done= true
  while(str_not_done) do
    screen:Update() -- Do everything in the callbacks.
    local evt= screen:GetEvent({wait= true, textok= true})
    if(evt.type == "keydown") then
      gs_keypress(evt.key)
    elseif(evt.type == "keyup") then
      gs_keyrelease(evt.key)
    end
  end
  text_ob= nil
  str_ob= nil
  str_pos= nil
  local ret= str
  str= nil -- Circumlocution necessary to clean up str.
  print(ret)
  return ret--]]
end
