if #arg < 2 then print("usage: loaffiler.lua logdirectory/ targetfile.lua (arguments to pass to targetfile.lua\nMake sure to clear logdirectory if you don't want stale stuff hanging around,\nand to include the trailing slash.") return end
local logdir = assert(table.remove(arg,1))
local target = assert(table.remove(arg,1))

local files = {}

local io_open = io.open -- SubCritical compatibility

local last_time,last_file,last_line
local didwarn
local function hook(why, line)
   local t = debug.getinfo(2, "S")
   local file = t.short_src:match("[^\\/]+$")
   if file == "loaffiler.lua" or not file or #file == 0 then return end
   if not files[file] then
      files[file] = {path=t.source}
   elseif not didwarn and files[file].path ~= t.source then
      didwarn = true
      io.stderr:write("Multiple files with the same name, confusing loaffiler! Output will be silly!\n")
   end
   local now = os.clock()
   if last_file then
      files[last_file][last_line] = (files[last_file][last_line] or 0) + (now - last_time)
   end
   last_time,last_file,last_line = now,file,line
end

local function go()
   debug.sethook(hook, "l")
   return dofile(target)
end
local s,e = xpcall(go,debug.traceback)
debug.sethook()
if not s then
   io.stderr:write(e.."\n")
end

for file,wut in pairs(files) do
   local lines = {}
   if wut.path:sub(1,1) == "@" then
      local f = io_open(wut.path:sub(2,-1),"r")
      if not f then
         io.stderr:write("Code for "..wut.path.." not found!")
      else
         repeat
            line = f:read()
            lines[#lines+1] = line
         until not line
         f:close()
      end
   end
   local t = {}
   for k,v in pairs(wut) do
      if type(k) == "number" then
         t[#t+1] = {lineno=k,("%10.2f %s"):format(v,lines[k] or ("(line "..k.."))"))}
      end
   end
   for n=1,#lines do
      if not wut[n] then
         t[#t+1] = {lineno=n,("           %s"):format(lines[n] or ("(line "..n.."))"))}
      end
   end
   table.sort(t, function(a,b) return a.lineno < b.lineno end)
   for n=1,#t do
      t[n] = t[n][1]
   end
   local f = assert(io_open(logdir..file,"w"))
   f:write(table.concat(t,"\n").."\n")
   f:close()
end
