--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

require "tile"

dir_name_to_index= {up= 0, top= 0, right= 1, down= 2, bottom= 2, left= 3}
dir_index_to_name= {[0]= "top", [1]= "right", [2]= "bottom", [3]= "left"}


function map_init()
  -- require "teh_map"
  brick_grc= png:Load(SCPath("bricks.png"))
  -- KLUDGE!
  --[[
  for n= 2, 10, 1 do
    _G["wall_h" .. n .. "_grc"]= png:Load(SCPath("wall_h" .. n .. ".png"))
    _G["wall_v" .. n .. "_grc"]= png:Load(SCPath("wall_v" .. n .. ".png"))
  end
]]
end

function two_d_to_room(array)
  local arraysize= array.size
  local arraysizex= arraysize.x
  local arraysizey= arraysize.y
  local ret= {walls= {}, places= {}, shots= {}, avatars= {}, size= {x= arraysizex, y= arraysizey}}
  --print("two_d_to_room start")
  for y=1, arraysizey, 1 do
    for x=1, arraysizex, 1 do
      if(array[y][x] == nil) then
        --print("two_d_to_room overstepped array bounds!  " .. x .. ", " .. y)
        return nil
      end
      if(array[y][x] ~= 0) then
        local px, py= ((x-1)*32) + 16, ((y-1)*32) + 16
        local wid= pos_to_wid(arraysizex, px, py)
        local new_wall
        local l, t, r, b= false, false, false, false
        if(x > 1) then l= (array[y][x-1] == 0) end
        if(y > 1) then t= (array[y-1][x] == 0) end
        if(x < arraysizex) then r= (array[y][x+1] == 0) end
        if(y < arraysizey) then b= (array[y+1][x] == 0) end
        new_wall= create_partially_edged_tile(px, py, array[y][x], l, t, r, b)
        ret.walls[wid]= new_wall
        --[[
        -- Look for nearby walls to join edges with.
        local hwalls_to_attach= {new_wall}
        local vwalls_to_attach= {new_wall}
        local ox= x
        while(x < arraysizex and array[y][x+1] ~= 0) do
          x= x + 1
          px, py= ((x-1)*32) + 16, ((y-1)*32) + 16
          new_wall= create_tile(px, py, array[y][x])
          wid= pos_to_wid(arraysizex, px, py)
          --print("  Placing wall at " .. px .. ", " .. py .. " at wid " .. wid)
          ret.walls[wid]= new_wall
          hwalls_to_attach[#hwalls_to_attach+1]= new_wall
        end
        attach_walls_h(hwalls_to_attach)
        local ny
        for ny= y, arraysizey, 1 do
          if(array[ny][ox] == 0) then
            break
          end
          px, py= ((ox-1)*32) + 16, ((ny-1)*32) + 16
          new_wall= create_tile(px, py, array[ny][ox])
          wid= pos_to_wid(arraysizex, px, py)
          ret.walls[wid]= new_wall
          --print("  Placing wall at " .. px .. ", " .. py .. " at wid " .. wid)
          vwalls_to_attach[#hwalls_to_attach+1]= new_wall
          -- Make sure we don't try to create another tile for this square.
          array[ny][ox]= 0
        end
        attach_walls_v(vwalls_to_attach)
        --]]
      end
    end
  end
  --print("two_d_to_room end")
  return ret
end

function attach_walls_h(walls)
  local xmin= walls[1].pos.x - walls[1].size.x
  local xmax= walls[1].pos.x + walls[1].size.x
  local ymin= walls[1].pos.y - walls[1].size.y
  local ymax= walls[1].pos.y + walls[1].size.y
  for name,wall in pairs(walls) do
    xmin= math.min(xmin, wall.pos.x - wall.size.x)
    xmax= math.max(xmax, wall.pos.x + wall.size.x)
  end
  local top= {a= sc.Construct("Vec2", xmin, ymin), b= sc.Construct("Vec2", xmax, ymin), normal= sc.Construct("Vec2", 0, -1)}
  local bottom= {a= sc.Construct("Vec2", xmin, ymax), b= sc.Construct("Vec2", xmax, ymax), normal= sc.Construct("Vec2", 0, 1)}
  for name,wall in pairs(walls) do
    if(wall.edges.left and wall.edges.left.a.x > xmin) then
      wall.edges.left= nil
    end
    if(wall.edges.right and wall.edges.right.a.x < xmax) then
      wall.edges.right= nil
    end
    wall.edges.top= top
    wall.edges.bottom= bottom
  end
end

function attach_walls_v(walls)
  local xmin= walls[1].pos.x - walls[1].size.x
  local xmax= walls[1].pos.x + walls[1].size.x
  local ymin= walls[1].pos.y - walls[1].size.y
  local ymax= walls[1].pos.y + walls[1].size.y
  for name,wall in pairs(walls) do
    ymin= math.min(ymin, wall.pos.y - wall.size.y)
    ymax= math.max(ymax, wall.pos.y + wall.size.y)
  end
  local left= {a= sc.Construct("Vec2", xmin, ymin), b= sc.Construct("Vec2", xmin, ymax), normal= sc.Construct("Vec2", -1, 0)}
  local right= {a= sc.Construct("Vec2", xmax, ymin), b= sc.Construct("Vec2", xmax, ymax), normal= sc.Construct("Vec2", 1, 0)}
  for name,wall in pairs(walls) do
    if(wall.edges.top and wall.edges.top.a.y > ymin) then
      wall.edges.top= nil
    end
    if(wall.edges.bottom and wall.edges.bottom.a.y < ymax) then
      wall.edges.bottom= nil
    end
    wall.edges.left= left
    wall.edges.right= right
  end
end

function room_print(foo)
  print("s:  " .. foo.size.x .. ", " .. foo.size.y)
  local n= 1
  while(foo.walls[n]) do
    --print("w" .. n .. ":  " .. foo.walls[n].x .. ", " .. foo.walls[n].y .. ", " .. foo.walls[n].w .. ", " .. foo.walls[n].h)
    print("w" .. n .. ":  " .. foo.walls[n].pos.x .. ", " .. foo.walls[n].pos.y, ", " .. foo.walls[n].size.x .. ", " ..
          foo.walls[n].size.y)
    n= n+1
  end
  n= 1
  while(foo.places[n]) do
    print("p" .. n .. ":  " .. foo.places[n].name .. ", " .. foo.places[n].x
     .. ", " .. foo.places[n].y .. ", " .. foo.places[n].w .. ", " .. foo.places[n].h)
    n= n+1
  end
  print("c:  l: (" .. foo.conns.left.side .. ", " ..
    foo.conns.left.rot .. ")  t: (" .. foo.conns.top.side .. ", " ..
    foo.conns.top.rot .. ")  r: (" .. foo.conns.right.side .. ", " ..
    foo.conns.right.rot .. ")  b: (" .. foo.conns.bottom.side .. ", " ..
    foo.conns.bottom.rot .. ")")
  print("\n")
end

function rot_point0(p, area)
  return p
end

-- 90 degrees clockwise
function rot_point1(p, area)
  return sc.Construct("Vec2", area.x - p.y, p.x)
end

-- 180 degrees
function rot_point2(p, area)
  return area - p
end

-- 90 degrees counterclockwise
function rot_point3(p, area)
  return sc.Construct("Vec2", p.y, area.y - p.x)
end

function rot_point_cen1(p)
  -- This may seem backwards, but it's because graphical positions have the y axis flipped.
  return sc.Construct("Vec2", -p.y, p.x)
end

function rot_point_cen2(p)
  return -p
end

function rot_point_cen3(p)
  -- This may seem backwards, but it's because graphical positions have the y axis flipped.
  return sc.Construct("Vec2", p.y, -p.x)
end

function rot_rect1(rect, size)
  return {x= (size.y - rect.y) - (rect.h), y= rect.x, w= rect.h, h= rect.w}
end

function rot_rect2(rect, size)
  return {x= (size.x - rect.x) - (rect.w), y= (size.y - rect.y) - (rect.h), w= rect.w,
    h= rect.h}
end

function rot_rect3(rect, size)
  return {x= rect.y, y= (size.x - rect.x) - (rect.w), w= rect.h, h= rect.w}
end

rot_point_funcs= {}
rot_point_funcs[0]= rot_point0
rot_point_funcs[1]= rot_point1
rot_point_funcs[2]= rot_point2
rot_point_funcs[3]= rot_point3
rot_point_cen_funcs= {}
rot_point_cen_funcs[0]= rot_point0
rot_point_cen_funcs[1]= rot_point_cen1
rot_point_cen_funcs[2]= rot_point_cen2
rot_point_cen_funcs[3]= rot_point_cen3
rot_point_inv_funcs= {}
rot_point_inv_funcs[0]= rot_point0
rot_point_inv_funcs[1]= rot_point3
rot_point_inv_funcs[2]= rot_point2
rot_point_inv_funcs[3]= rot_point1
rot_point_inv_cen_funcs= {}
rot_point_inv_cen_funcs[0]= rot_point0
rot_point_inv_cen_funcs[1]= rot_point_cen3
rot_point_inv_cen_funcs[2]= rot_point_cen2
rot_point_inv_cen_funcs[3]= rot_point_cen1

function build_minimap_from_complete_room(room, dest, dw, dh)
  local minigraph= dest
  minigraph:SetPrimitiveColor(0, 0, 0)
  minigraph:DrawRect(0, 0, dw, dh)
  local room_walls= room.walls
  local wall_points= {}
  for wid, wall in pairs(room_walls) do
    local wall_pos= wall.pos
    wall_points[#wall_points+1]= math.floor(wall_pos.x / (32))
    wall_points[#wall_points+1]= math.floor(wall_pos.y / (32))
  end
  minigraph:SetPrimitiveColor(1, 1, 1)
  minigraph:DrawPoints(SCUtil.CompileCoords(wall_points), 1)
end

function load_room(room)
  local room_file, err= io.open(SCPath("rooms/"..room..".celduin"), "r")
  if not room_file then
    print("Failed to open file to load room:  "..err)
    return
  end
  local checksum= room_file:read(4)
  local celduin= room_file:read("*a")
  if checksum ~= SCUtil.QuickChecksum(celduin) then
    print("Room "..room.." file failed checksum.")
    return
  end
  local rehyd_room= SCUtil.Rehydrate(celduin)
  active_rooms[room]= two_d_to_room(rehyd_room.array)
  active_rooms[room].places= rehyd_room.places
  for k, place in pairs(active_rooms[room].places) do
    place.pos= sc.Construct("Vec2", place.pos.x, place.pos.y)
    place.size= sc.Construct("Vec2", place.size.x, place.size.y)
  end
  active_rooms[room].conns= rehyd_room.conns
end

function unload_room(room)
  active_rooms[room]= nil
  collectgarbage("collect")
end
