--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

require "commodities"

function planets_init()
  commodities_init()
  planets= {}
  
  create_planet("Klyztam", "A heavily industrialized planet with little farmland.", "base_pic1.png")
  planets.Klyztam.comm_makers= {make_eq_fast, eat_food_med, eat_med_med}
  planets.Klyztam.commodities= {Equipment= comm_eq_low, Food= comm_food_med, Medical= comm_med_med}
  planets.Klyztam.connections.Glorsap= true
  planets.Klyztam.connections.Fhqwhgads= true
  
  create_planet("Glorsap", "A peaceful farming planet.", "base_pic2.png")
  planets.Glorsap.comm_makers= {make_eq_med, make_food_fast, make_med_slow, eat_eq_slow, eat_med_med}
  planets.Glorsap.commodities= {Equipment= comm_eq_med, Food= comm_food_low, Medical= comm_med_med}
  planets.Glorsap.connections.Klyztam= true
  
  create_planet("Fhqwhgads", "A strabge qsabet wetg few twads.", "base_pic3.png")
  planets.Fhqwhgads.comm_makers= {make_food_med, eat_food_med, create_maker(5, {{amount= 1, name= "Toads"}}, {}, 30)}
  planets.Fhqwhgads.commodities= {Food= comm_food_med, Toads= create_comm("Toads", 10, 500, 0.15)}
  planets.Fhqwhgads.connections.Klyztam= true
  planets.Fhqwhgads.connections.Nowhere= true

  create_planet("Twad_Savd", "A strabge qsabet wetg cavt twads.", "base_pic4.png")
  planets.Twad_Savd.comm_makers= {make_eq_fast, make_food_slow, make_med_med, create_maker(5, {}, {"Toads"}, 30),
                                  eat_eq_med, eat_food_med, eat_med_slow}
  planets.Twad_Savd.commodities= {Equipment= comm_eq_low, Food= comm_food_high, Medical= comm_med_med, Toads= create_comm("Toads", 100, 10, 0.09)}
  planets.Twad_Savd.connections.Klyztam= true
  planets.Twad_Savd.connections.Nowhere= true

  create_planet("Nowhere", "Nothing important ever has or ever will happen here.", "base_pic5.png")
  planets.Nowhere.comm_makers= {make_eq_med, make_food_med, create_maker(5, {}, {"Tourists"}, 15), eat_eq_med, eat_food_fast}
  planets.Nowhere.commodities= {Equipment= comm_eq_med, Food= comm_food_med, Tourists= create_comm("Tourists", 50, -5, 0.06)}
  planets.Nowhere.connections.Fhqwhgads= true
  planets.Nowhere.connections.AEON= true

  create_planet("AEON", "The Airport East Of Nowhere.  Offering flights to anywhere in the known universe.", "base_pic6.png")
  planets.AEON.comm_makers= {make_eq_slow, create_maker(5, {{amount= 1, name= "Tourists"}}, {}, 15), eat_eq_med,
                             create_maker(10, {{amount= 1, name= "Food"}}, {}, 30)}
  planets.AEON.commodities= {Equipment= comm_eq_high, Food= create_comm("Food", 0, 50, 0.03), Tourists= create_comm("Tourists", 5, 10, 0.06)}
  planets.AEON.connections.Klyztam= true
  planets.AEON.connections.Glorsap= true
  planets.AEON.connections.Fhqwhgads= true
  planets.AEON.connections.Nowhere= true
  planets.AEON.connections.AEON= true
  planets.AEON.connections.Twad_Savd= true
  planets.AEON.connections["NGC-1"]= true

  --[[
  create_planet("NGC-1", "An empty system with only a waystation for travelers.", "base_pic2.png")
  planets["NGC-1"].connections.AEON= true
  planets["NGC-1"].connections["NGC-2"]= true

  create_planet("NGC-2", "An empty system with only a waystation for travelers.", "base_pic2.png")
  planets["NGC-2"].connections["NGC-1"]= true
  planets["NGC-2"].connections["NGC-3"]= true

  create_planet("NGC-3", "An empty system with only a waystation for travelers.", "base_pic2.png")
  planets["NGC-3"].connections["NGC-2"]= true
  planets["NGC-3"].connections["NGC-4"]= true

  create_planet("NGC-4", "An empty system with only a waystation for travelers.", "base_pic2.png")
  planets["NGC-4"].connections["NGC-3"]= true
  planets["NGC-4"].connections["NGC-5"]= true

  create_planet("NGC-5", "An empty system with only a waystation for travelers.", "base_pic2.png")
  planets["NGC-5"].connections["NGC-4"]= true
  planets["NGC-5"].connections["NGC-6"]= true

  create_planet("NGC-6", "An empty system with only a waystation for travelers.", "base_pic2.png")
  planets["NGC-6"].connections["NGC-5"]= true
  planets["NGC-6"].connections["NGC-7"]= true

  create_planet("NGC-7", "An empty system with only a waystation for travelers.", "base_pic2.png")
  planets["NGC-7"].connections["NGC-6"]= true
  planets["NGC-7"].connections.Kykinku= true

  create_planet("Kykinku",
    "The biggest tourist trap you've seen.  Everything costs more than it's worth.\n" ..
    " So devoted to tourism that a set of subsidiary planets is needed to provide it with\n" ..
    " basic necessities.", "base_pic2.png")
  planets.Kykinku.comm_makers= {create_maker(10, {{amount= 1, name= "Tourists"}}, {}, 15),
                                create_maker(10, {{amount= 1, name= "Food"}}, {}, 15),
                                create_maker(10, {{amount= 1, name= "Equipment"}}, {}, 15),
                                create_maker(10, {{amount= 1, name= "Medical"}}, {}, 15),
                                create_maker(5, {}, {"Souvenirs"}, 30)}
  planets.Kykinku.connections["NGC-7"]= true
  planets.Kykinku.connections.Slevetz= true
  planets.Kykinku.connections.Mezork= true
  planets.Kykinku.connections.Florm= true

  create_planet("Slevetz",
    "This planet is so industrialized, it just looks like a steel globe from space.\n" ..
    "It must be heavily mined out, your sensors tell you that the overall density is\n ~0.9," ..
    " and it's much larger than most planets.  It supplies Equipment to Kykinku.", "base_pic2.png")
  planets.Slevetz.comm_makers= {create_maker(10, {}, {"Equipment"}, 15), make_med_slow, make_food_slow, eat_med_med, eat_food_med}
  planets.Slevetz.connections.Kykinku= true
  planets.Slevetz.connections.Mezork= true
  planets.Slevetz.connections.Florm= true

  create_planet("Mezork",
    "A vast complex of hospitals and medical research buildings covers most of this\n planet.", "base_pic2.png")
  planets.Mezork.comm_makers= {create_maker(10, {}, {"Medical"}, 15), make_eq_slow, make_food_slow, eat_eq_med, eat_food_med}
  planets.Mezork.connections.Kykinku= true
  planets.Mezork.connections.Slevetz= true
  planets.Mezork.connections.Florm= true

  create_planet("Florm",
    "At first, this planet seems uninhabited, but then you realize that it's one vast\n farm.", "base_pic2.png")
  planets.Florm.comm_makers= {create_maker(10, {}, {"Food"}, 15), make_med_fast, make_eq_med, eat_med_fast, eat_eq_med}
  planets.Florm.connections.Kykinku= true
  planets.Florm.connections.Slevetz= true
  planets.Florm.connections.Mezork= true

  planets.Quientish= {name= "Quientish", desc= "Home planet of the Quientish, galactic slave traders.",
  img_file_name= "quientish.png", commodities= {}, connections= {}}
]]
  print("Planets inited")
  --table_print(planets)
end

function create_planet(name, desc, img_name)
  planets[name]= {comm_makers= {}, commodities= {}, connections= {}, name= name, desc= desc, img_file_name= img_name}
end

function pl_update(this)
  if not this.comm_makers then return end
  for id, maker in pairs(this.comm_makers) do
    if gt_rticks >= maker.next_make_time then
      run_comm_maker(maker, this)
    end
  end
  for name, comm in pairs(this.commodities) do
    if comm.amount ~= comm.amount_last_tick then
      comm_set_price(comm, this)
      comm.amount_last_tick= comm.amount
    end
  end
end
