--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

--firea_coords= {}
--firea_indices= {}
--fireb_coords= {}
--fireb_indices= {}

function gui_player_init()
  print("Creating player")
  next_acw_ori= {up= "left", left= "down", down= "right", right= "up"}
  next_cw_ori= {up= "right", left= "up", down= "left", right= "down"}
  rev_dir= {right= "left", left= "right"}
  player= create_basic_ava("Kyzentun", "tank1_main_body_sprites.png", 1, sc.Construct("Vec2", 1792, 1792),
                           {1, 1, 1})
--  table_print(player)
  dirs= {"right", "left"}
  ors= {"up", "right", "down", "left"}
  ori_to_ind= {up= 0, right= 1, down= 2, left= 3}
  mots= {"still", "moving", "inside_corner", "outside_corner", "jumping", "falling",
    "spin_jump_cw", "spin_jump_acw", "turning_in_air", "turning"}
  fill_in_anim_info(player)
  print("player inited")
  create_basic_ais()
  actor= {
    tickly_actions= function(this)
      --profile_section_start("av_cont_jump")
      actor.continue_jump(this)
      --profile_section_end("av_cont_jump")
      --profile_section_start("av_update_anim")
      actor.update_anim(this)
      --profile_section_end("av_update_anim")
      if(not fly) then
        --profile_section_start("av_grav")
        actor.apply_gravity(this, .25)
        --profile_section_end("av_grav")
      end
      --profile_section_start("av_rr")
      actor.recharge_and_repair(this)
      --profile_section_end("av_rr")
      --profile_section_start("av_try_move")
      actor.try_to_move(this)
      --profile_section_end("av_try_move")
      --profile_section_start("av_try_aim")
      actor.try_to_aim(this)
      --profile_section_end("av_try_aim")
      --profile_section_start("av_try_fire")
      actor.try_to_fire(this)
      --profile_section_end("av_try_fire")
      --profile_section_start("av_update_speed")
      actor.update_speed(this)
      --profile_section_end("av_update_speed")
    end,
    add_rot= function(this, dir, rot)
    -- An explanation of cur_rot and orienation
    -- cur_rot is how the world appears to be rotated to the avatar.  This is also an indicator of the direction of gravity
    -- for the avatar.
    -- orienation is how the avatar is oriented with respect to the world.  This is the directin the avatar's "up" vector points.
      if(rot == 0) then
        --io.write("  Changing pos from " .. this.pos.x .. ", " .. this.pos.y)
        if(dir == "left") then
          -- Since the avatar is exiting through the left side, they would logically now be on the right side.
          -- But since the room they are entering into is rotated, we need to perform the inverse rotation to find where they actually enter at.
          this.pos= sc.Construct("Vec2", world_max.x, this.pos.y)
        elseif(dir == "right") then
          this.pos= sc.Construct("Vec2", 0, this.pos.y)
        elseif(dir == "top") then
          this.pos= sc.Construct("Vec2", this.pos.x, world_max.y)
        elseif(dir == "bottom") then
          this.pos= sc.Construct("Vec2", this.pos.x, 0)
        end
        --print(" to " .. this.pos.x .. ", " .. this.pos.y)
        return
      end
      local new_rot= math.mod(this.cur_rot + rot, 4)
      --print("Changing rot by " .. rot .. " from " .. this.cur_rot .. " to " .. new_rot)
      -- Adding in the rotation change from the room change has three effects.
      -- 1.  The position of the avatar changes to correspond to the side they are entering through
      -- 1.  The oriendation of the avatar has to change to appear the same since it will now be rotated with the room when blitting.
      -- 3.  The speed of the avatar changes to keep the avatar moving in the same direction relative to the player.
      -- 3 can be dealt with by itself, but 1 and 2 interact because they both adjust the physical position of the avatar.
      -- 2 adjusts the position to make the avatar appear to stay in the same place when its orientation changes, even though the avatar's g_offset changes.
      -- To keep these two from causing a drift when moving between rooms, 2 is split into two parts, and 1 is performed between them.
      if(this.anim_info.status.orientation == "up") then
        this.pos= sc.Construct("Vec2", this.pos.x, this.pos.y - 16)
        if(rot == 1) then
          this.anim_info.status.orientation= "left"
          this.size= sc.Construct("Vec2", this.size.y, this.size.x)
        elseif(rot == 2) then
          this.anim_info.status.orientation= "down"
        elseif(rot == 3) then
          this.anim_info.status.orientation= "right"
          this.size= sc.Construct("Vec2", this.size.y, this.size.x)
        end
      elseif(this.anim_info.status.orientation == "right") then
        this.pos= sc.Construct("Vec2", this.pos.x + 16, this.pos.y)
        if(rot == 1) then
          this.anim_info.status.orientation= "up"
          this.size= sc.Construct("Vec2", this.size.y, this.size.x)
        elseif(rot == 2) then
          this.anim_info.status.orientation= "left"
        elseif(rot == 3) then
          this.anim_info.status.orientation= "down"
          this.size= sc.Construct("Vec2", this.size.y, this.size.x)
        end
      elseif(this.anim_info.status.orientation == "down") then
        this.pos= sc.Construct("Vec2", this.pos.x, this.pos.y + 16)
        if(rot == 1) then
          this.anim_info.status.orientation= "right"
          this.size= sc.Construct("Vec2", this.size.y, this.size.x)
        elseif(rot == 2) then
          this.anim_info.status.orientation= "up"
        elseif(rot == 3) then
          this.anim_info.status.orientation= "left"
          this.size= sc.Construct("Vec2", this.size.y, this.size.x)
        end
      elseif(this.anim_info.status.orientation == "left") then
        this.pos= sc.Construct("Vec2", this.pos.x - 16, this.pos.y)
        if(rot == 1) then
          this.anim_info.status.orientation= "down"
          this.size= sc.Construct("Vec2", this.size.y, this.size.x)
        elseif(rot == 2) then
          this.anim_info.status.orientation= "right"
        elseif(rot == 3) then
          this.anim_info.status.orientation= "up"
          this.size= sc.Construct("Vec2", this.size.y, this.size.x)
        end
      end
      --io.write("  Changing pos from " .. this.pos.x .. ", " .. this.pos.y)
      if(dir == "left") then
        -- Since the avatar is exiting through the left side, they would logically now be on the right side.
        -- But since the room they are entering into is rotated, we need to perform the inverse rotation to find where they actually enter at.
        this.pos= rot_point_inv_funcs[rot](sc.Construct("Vec2", this.pos.x + world_max.x, this.pos.y), world_max)
      elseif(dir == "right") then
        this.pos= rot_point_inv_funcs[rot](sc.Construct("Vec2", this.pos.x - world_max.x, this.pos.y), world_max)
      elseif(dir == "top") then
        this.pos= rot_point_inv_funcs[rot](sc.Construct("Vec2", this.pos.x, this.pos.y + world_max.y), world_max)
      elseif(dir == "bottom") then
        this.pos= rot_point_inv_funcs[rot](sc.Construct("Vec2", this.pos.x, this.pos.y - world_max.y), world_max)
      end
      --print(" to " .. this.pos.x .. ", " .. this.pos.y)
      -- Remember that orientation was changed by the above elseif chain.
      if(this.anim_info.status.orientation == "up") then
        this.pos= sc.Construct("Vec2", this.pos.x, this.pos.y + 16)
        this.g_offset= sc.Construct("Vec2", 0, -32)
      elseif(this.anim_info.status.orientation == "right") then
        this.pos= sc.Construct("Vec2", this.pos.x - 16, this.pos.y)
        this.g_offset= sc.Construct("Vec2", 0, 0)
      elseif(this.anim_info.status.orientation == "down") then
        this.pos= sc.Construct("Vec2", this.pos.x, this.pos.y - 16)
        this.g_offset= sc.Construct("Vec2", 0, 0)
      elseif(this.anim_info.status.orientation == "left") then
        this.pos= sc.Construct("Vec2", this.pos.x + 16, this.pos.y)
        this.g_offset= sc.Construct("Vec2", -32, 0)
      end
      --io.write("Changing speed from " .. this.speed.x .. ", " .. this.speed.y)
      this.speed= rot_point_inv_cen_funcs[rot](this.speed)
      --print(" to " .. this.speed.x .. ", " .. this.speed.y)
      this.speed_adjust= rot_point_inv_cen_funcs[rot](this.speed_adjust)
      this.cur_rot= new_rot
      actor.update_image_window(this)
    end,
    aim_down= function(this)
      for id, gun in pairs(this.ship.primary_guns) do
        gun.desired_aim= force_aim_in_range(math.pi * .5, gun.aim_min, gun.aim_max)
      end
    end,
    aim_ahead= function(this)
      for id, gun in pairs(this.ship.primary_guns) do
        gun.desired_aim= force_aim_in_range(0, gun.aim_min, gun.aim_max)
      end
    end,
    aim_mid_range= function(this)
      for id, gun in pairs(this.ship.primary_guns) do
        local amin, amax= gun.aim_min, gun.aim_max
        if amin > amax then amin= amin - (math.pi * 2) end
        gun.desired_aim= (amin + amax) / 2
      end
    end,
    aim_up= function(this)
      for id, gun in pairs(this.ship.primary_guns) do
        gun.desired_aim= force_aim_in_range(math.pi * 1.5, gun.aim_min, gun.aim_max)
      end
    end,
    aim_arbitrary= function(this, aim)
      local avasta= this.anim_info.status
      local relative_aim= dir_aim_flip[avasta.dir](aim + ori_to_aim_unoffset_table[avasta.orientation])
      for id, gun in pairs(this.ship.primary_guns) do
        gun.desired_aim= force_aim_in_range(relative_aim, gun.aim_min, gun.aim_max)
      end
    end,
    apply_gravity= function(this, grav_strength)
      -- Only apply gravity if there is not something below us.
      
      -- Fudge below_rect so that we won't think a tile that is off to the side is below.
      this.below_rect= {size= sc.Construct("Vec2", this.size.x - 2, 2)}
      local grav_vec
      if(this.cur_rot == 0) then
        this.below_rect.pos= sc.Construct("Vec2", this.pos.x, this.pos.y + this.size.y)
        grav_vec= sc.Construct("Vec2", 0, grav_strength)
      elseif(this.cur_rot == 1) then
        this.below_rect.pos= sc.Construct("Vec2", this.pos.x + this.size.x, this.pos.y)
        this.below_rect.size= sc.Construct("Vec2", 2, this.size.y -2)
        grav_vec= sc.Construct("Vec2", grav_strength, 0)
      elseif(this.cur_rot == 2) then
        this.below_rect.pos= sc.Construct("Vec2", this.pos.x, this.pos.y - this.size.y)
        grav_vec= sc.Construct("Vec2", 0, -grav_strength)
      elseif(this.cur_rot == 3) then
        this.below_rect.pos= sc.Construct("Vec2", this.pos.x - this.size.x, this.pos.y)
        this.below_rect.size= sc.Construct("Vec2", 2, this.size.y -2)
        grav_vec= sc.Construct("Vec2", -grav_strength, 0)
      end
      local should_apply= not check_wall_in_rect(this.cur_room, this.below_rect)
      if(should_apply) then
        if(rev_grav) then
          this.speed= this.speed - grav_vec
        else
          this.speed= this.speed + grav_vec
        end
      else
        -- Do not end the jump if the avatar is moving against the pull of gravity.
        if(this.speed .. grav_vec > 0) then
          actor.end_jump(this)
        end
      end
    end,
    apply_motion= function(this, v)
                    this.pos= this.pos + v
    end,
    apply_vector= function(this, v)
                    this.speed= this.speed + v
    end,
    negate_speed= function(this, side)
      if(size == "left" or side == "right") then
        this.speed= sc.Construct("Vec2", 0, this.speed.y)
      end
      if(side == "top" or side == "bottom") then
        this.speed= sc.Construct("Vec2", this.speed.x, 0)
      end
    end,
    recharge_and_repair= function(this)
                           this.ship.shield= this.ship.shield + this.ship.shield_recharge_rate
                           this.ship.armor= this.ship.armor + this.ship.armor_repair_rate
                           if this.ship.shield > this.ship.shield_max then this.ship.shield= this.ship.shield_max end
                           if this.ship.armor > this.ship.armor_max then this.ship.armor= this.ship.armor_max end
                         end,
    stop_motion= function(this, which)
      if(which == nil or which == "" or which == "xy") then
        this.speed= sc.Construct("Vec2")
      elseif(which == "x") then
        this.speed= sc.Construct("Vec2", 0, this.speed.y)
      elseif(which == "y") then
        this.speed= sc.Construct("Vec2", this.speed.x, 0)
      end
    end,
    start_firing= function(this)
      this.general_status.fire_main_down= true
    end,
    stop_firing= function(this)
      this.general_status.fire_main_down= nil
    end,
    try_to_aim= function(this)
      for id, gun in pairs(this.ship.primary_guns) do
        if gun.aim ~= gun.desired_aim then
          -- These calculations are here to make sure the gun turns in the right direction when crossing the 0 to 2pi boundary.
          local norm_diff= gun.desired_aim - gun.aim
          local plus_diff= (gun.desired_aim + (math.pi * 2)) - gun.aim
          local minus_diff= (gun.desired_aim - (math.pi * 2)) - gun.aim
          local min_diff= min_abs3(norm_diff, plus_diff, minus_diff)
          if math.abs(min_diff) <= gun.aim_speed then
            gun.aim= gun.desired_aim
          else
            if min_diff > 0 then
              gun.aim= gun.aim + gun.aim_speed
            else
              gun.aim= gun.aim - gun.aim_speed
            end
            gun.aim= wrap_aim(gun.aim)
          end
        end
        --local avasta= this.anim_info.status
        --local avadir= avasta.dir
        --local avaori= avasta.orientation
        --local real_pos= this.pos + rot_point_cen_funcs[ori_to_ind[avaori]](dir_offset_flip[avadir](gun.offset))
        --local firing_vec= calc_firing_vec(avadir, avaori, gun.aim, 0, gun.muzzle_velocity, this.speed)
        --add_point_to_coords(firea_coords, firea_indices, real_pos)
        --add_point_to_coords(firea_coords, firea_indices, real_pos + firing_vec)
      end
    end,
    try_to_fire= function(this)
      if this.general_status.fire_main_down then
        for id,gun in pairs(this.ship.primary_guns) do
          if gun.ammo_cap == -1 or gun.ammo_left > 0 then
            if gun.next_fire_time <= gt_rticks then
              gun.firing_func(gun, this, this.pos, this.speed, this.name, active_rooms[this.cur_room])
            end
          end
        end
      end
    end,
    try_to_move= function(this)
      local sp= 6
      if(this.general_status.move_left_down) then
        if(not this.general_status.move_right_down) then
          -- This rotates the speed around the origin to get the speed relative to the tank.
          local relative_speed= rot_point_cen_funcs[this.cur_rot](this.speed)
          if(relative_speed.x > -sp) then
            local newx= relative_speed.x - sp
            if(newx < -sp) then newx= -sp end
            relative_speed= sc.Construct("Vec2", newx, relative_speed.y)
            this.speed= rot_point_inv_cen_funcs[this.cur_rot](relative_speed)
          end
        end
      elseif(this.general_status.move_right_down) then
        local relative_speed= rot_point_cen_funcs[this.cur_rot](this.speed)
        if(relative_speed.x < sp) then
          local newx= relative_speed.x + sp
          if(newx > sp) then newx= sp end
          relative_speed= sc.Construct("Vec2", newx, relative_speed.y)
          this.speed= rot_point_inv_cen_funcs[this.cur_rot](relative_speed)
        end
      end
    end,
    update_anim= function(this)
--      print(gt_rticks .. ", " .. this.anim_info.frames[this.anim_info.status.dir]
  --      [this.anim_info.status.orientation][this.anim_info.status.motion].ticks)
      if(math.mod(gt_rticks, this.anim_info.frames[this.anim_info.status.dir]
        [this.anim_info.status.orientation][this.anim_info.status.motion].ticks) == 0) then
        this.anim_info.status.f_num= this.anim_info.status.f_num + 1
        if(this.anim_info.status.f_num > this.anim_info.frames[this.anim_info.status.dir]
          [this.anim_info.status.orientation][this.anim_info.status.motion].num) then
          this.anim_info.status.f_num= 1
          if(this.anim_info.status.motion == "inside_corner") then
            this.anim_info.status.motion= "moving"
            this.anim_info.status.orientation= next_acw_ori[this.anim_info.status.orientation]
          elseif(this.anim_info.status.motion == "outside_corner") then
            this.anim_info.status.motion= "moving"
            this.anim_info.status.orientation= next_cw_ori[this.anim_info.status.orientation]
          elseif(this.anim_info.status.motion == "jumping") then
            this.anim_info.status.motion= "falling"
          elseif(this.anim_info.status.motion == "turning_in_air") then
            this.anim_info.status.motion= "falling"
            this.anim_info.status.dir= rev_dir[this.anim_info.status.dir]
          elseif(this.anim_info.status.motion == "turning") then
            this.anim_info.status.motion= "moving"
            this.anim_info.status.dir= rev_dir[this.anim_info.status.dir]
          end
        end
        actor.update_image_window(this)
      end
    end,
    update_image_window= function(this)
      local r= this.anim_info.frames[this.anim_info.status.dir]
      [this.anim_info.status.orientation][this.anim_info.status.motion].window[this.anim_info.status.f_num]
--      rect_print(r)
      this.blit_image= sc.Construct("FakeDrawable", this.image, r.x, r.y, r.w, r.h)
      this.bg_frisket= SCUtil.MakeFrisketFromAlpha(this.blit_image)
      this.color_frisket= SCUtil.MakeFrisketFromGrayscale(this.blit_image)
      --print("bg_frisket and color_frisket made for " .. this.name)
    end,
    update_pos= function(this)
      this.old_pos= this.pos
      this.g_pos= this.pos + this.g_offset
    end,
    update_speed= function(this)
      this.speed= this.speed + this.speed_adjust
      this.speed_adjust= sc.Construct("Vec2")
    end,
    check_in_room= function(this)
      if(this.pos.x < 0) then
        return {ava= this, room= this.cur_room, dir= "left"}
      elseif(this.pos.y < 0) then
        return {ava= this, room= this.cur_room, dir= "top"}
      elseif(this.pos.x > world_max.x) then
        return {ava= this, room= this.cur_room, dir= "right"}
      elseif(this.pos.y > world_max.y) then
        return {ava= this, room= this.cur_room, dir= "bottom"}
      end
      return nil
    end,
    collision_rect= function(this, which)
      if(which == nil or which == "") then
        return {x= this.pos.x, y= this.pos.y, w= this.ship.w, h= this.ship.h}
      elseif(which == "xy") then
        return {x= this.pos.x + this.speed.x, y= this.pos.y + this.speed.y, w= this.ship.w,
          h= this.ship.h}
      elseif(which == "x") then
        return {x= this.pos.x + this.speed.x, y= this.pos.y, w= this.ship.w, h= this.ship.h}
      elseif(which == "y") then
        return {x= this.pos.x, y= this.pos.y + this.speed.y, w= this.ship.w, h= this.ship.h}
      end
      return nil
    end,
    continue_jump= function(this)
      if(this.in_jump and this.in_jump < this.jump_max) then
        local jump_vec= rot_point_inv_cen_funcs[this.cur_rot](sc.Construct("Vec2", 0, -.5))
        this.speed= this.speed + jump_vec
        this.in_jump= this.in_jump + 1
      end
    end,
    end_jump= function(this)
      this.in_jump= nil
      this.anim_info.status.motion= (xor(this.general_status.move_left_down, this.general_status.move_right_down))
        and "moving" or "still"
    end,
    jump= function(this)
      if(this.in_jump == nil) then
        local jump_vec= rot_point_inv_cen_funcs[this.cur_rot](sc.Construct("Vec2", 0, -.9))
        this.speed= this.speed + jump_vec
        this.in_jump= 1
        this.anim_info.status.motion= "jumping"
      end
    end
    }
  write_moves_funcs("left", "right")
  write_moves_funcs("right", "left")
end

function write_moves_funcs(foo, afoo)
  actor["start_" .. foo .. "_move"]= function(this)
    this.general_status["move_" .. foo .. "_down"]= true
    if(this.anim_info.status.motion == "still" or this.anim_info.status.motion == "moving") then
      this.anim_info.status.dir= foo
      if(this.general_status["move_" .. afoo .. "_down"]) then
        this.anim_info.status.motion= "still"
        if(this.cur_rot == 0 or this.cur_rot == 2) then
          this.speed.x= 0
        elseif(this.cur_rot == 1 or this.cur_rot == 3) then
          this.speed.y= 0
        end
      else
        this.anim_info.status.motion= "moving"
      end
    end
  end
  actor["end_" .. foo .. "_move"]= function(this)
    this.general_status["move_" .. foo .. "_down"]= nil
    if(this.general_status["move_" .. afoo .. "_down"]) then
      actor["start_" .. afoo .. "_move"](this)
    else
      this.anim_info.status.motion= "still"
      if(this.cur_rot == 0 or this.cur_rot == 2) then
        this.speed.x= 0
      elseif(this.cur_rot == 1 or this.cur_rot == 3) then
        this.speed.y= 0
      end
    end
  end
end

function anim_info_print(aif)
  print("aif start")
  local d= 1
  while(dirs[d]) do
    print(dirs[d])
    local o= 1
    while(ors[o]) do
      print("  " .. ors[o])
      local m= 1
      while(mots[m]) do
        print("    " .. mots[m])
        for w= 1, aif[dirs[d]][ors[o]][mots[m]].num, 1 do
          rect_print(aif[dirs[d]][ors[o]][mots[m]].window[w], "      ")
        end
        m= m+1
      end
      o= o+1
    end
    d= d+1
  end
  print("aif end")
end

function fill_in_anim_info(avatar)
  local d= 1
  while(dirs[d]) do
    avatar.anim_info.frames[dirs[d]]= {}
    local o= 1
    while(ors[o]) do
      avatar.anim_info.frames[dirs[d]][ors[o]]= {}
      local m= 1
      while(mots[m]) do
        avatar.anim_info.frames[dirs[d]][ors[o]][mots[m]]= {num= 3, ticks= 2, window= {
          {x= ((d-1) * 640) + ((m-1) * 64), y= ((o-1) * 192), w= 64, h= 64},
          {x= ((d-1) * 640) + ((m-1) * 64), y= ((o-1) * 192) + 64, w= 64, h= 64},
          {x= ((d-1) * 640) + ((m-1) * 64), y= ((o-1) * 192) + 128, w= 64, h= 64}
          }}
        m= m+1
      end
      o= o+1
    end
    d= d+1
  end
  d= 1
  while(dirs[d]) do
    local o= 1
    while(ors[o]) do
      avatar.anim_info.frames[dirs[d]][ors[o]].jumping.ticks= 20
      o= o+1
    end
    d= d+1
  end
--  anim_info_print(avatar.anim_info.frames)
end

-- Moved out of the function so it isn't recreated every time the function is called.
local courses_of_action= {"nothing", "jump", "move left", "jump left", "move right", "jump right"}

function basic_ai_think(this, ticks, room)
  if(ticks > this.next_think_tick) then
    local prev_coa= this.course_of_action
    if(this.move_randomly) then
      this.course_of_action= courses_of_action[math.random(#courses_of_action)]
    elseif(this.leaving_place) then
      local jom= "jump"
      if(math.random(3) ~= 3) then
        jom= "move"
      end
      if(this.cur_rot == 0) then
        if(this.pos.x < world_max.x / 2) then
          this.course_of_action= jom .. " left"
        else
          this.course_of_action= jom .. " right"
        end
      elseif(this.cur_rot == 1) then
        if(this.pos.y < world_max.y / 2) then
          this.course_of_action= jom .. " right"
        else
          this.course_of_action= jom .. " left"
        end
      elseif(this.cur_rot == 2) then
        if(this.pos.x < world_max.x / 2) then
          this.course_of_action= jom .. " right"
        else
          this.course_of_action= jom .. " left"
        end
      elseif(this.cur_rot == 3) then
        if(this.pos.y < world_max.y / 2) then
          this.course_of_action= jom .. " left"
        else
          this.course_of_action= jom .. " right"
        end
      end
    elseif(this.predisposition) then
      this.course_of_action= this.predisposition
      this.predisposition= nil
    else
      local place= room.places[1]
      if(place) then
        local jom= "move"
        local dir= " left"
        if(this.cur_rot == 0) then
          if(this.pos.x < place.pos.x) then
            dir= " right"
          end
          if(this.pos.y > place.pos.y) then
            jom= "jump"
          end
        elseif(this.cur_rot == 1) then
          if(this.pos.y > place.pos.y) then
            dir= " right"
          end
          if(this.pos.x > place.pos.x) then
            jom= "jump"
          end
        elseif(this.cur_rot == 2) then
          if(this.pos.x > place.pos.x) then
            dir= " right"
          end
          if(this.pos.y < place.pos.y) then
            jom= "jump"
          end
        elseif(this.cur_rot == 3) then
          if(this.pos.y < place.pos.y) then
            dir= " right"
          end
          if(this.pos.x < place.pos.x) then
            jom= "jump"
          end
        end
        this.course_of_action= jom .. dir
      end
    end
    this.next_think_tick= ticks + 60
    local coa= this.course_of_action -- Shortened to avoid hash table lookups.
    --print("AI " .. this.name .. " decided to do " .. coa)
    if(prev_coa == "move right" or prev_coa == "jump right" or prev_coa == "hop right") then
      --print("Ending right move.")
      actor.end_right_move(this)
    end
    if(prev_coa == "move left" or prev_coa == "jump left" or prev_coa == "hop left") then
      --print("Ending left move.")
      actor.end_left_move(this)
    end
    if(coa == "nothing") then
    end
    if (prev_coa == "jump" or prev_coa == "jump left" or prev_coa == "jump right" or prev_coa == "hop left" or prev_coa == "hop right")
      and not (coa == "jump" or coa == "jump left" or coa == "jump right") then
      --print("Jumping.")
      this.in_jump= this.jump_max
    end
    if(coa == "jump" or coa == "jump left" or coa == "jump right") then
      --print("Jumping.")
      actor.jump(this)
    end
    if(coa == "hop" or coa == "hop left" or coa == "hop right") then
      actor.jump(this)
      this.next_think_tick= ticks + 15 -- Modify next_think_tick to make the AI rethink (and probably stop) the hop in 15 ticks.
    end
    if(coa == "move left" or coa == "jump left" or coa == "hop left") then
      --print("Starting left move.")
      actor.start_left_move(this)
    end
    if(coa == "move right" or coa == "jump right" or coa == "hop right") then
      --print("Starting right move.")
      actor.start_right_move(this)
    end
    if this.enemy then
      --print(this.name .. " has an enemy named " .. this.enemy)
      if room.avatars[this.enemy] then
        local enemy= room.avatars[this.enemy]
        local aim_vec= enemy.pos - this.pos
        local aim_dir= wrap_aim(math.atan2(aim_vec.y, aim_vec.x))
        actor.aim_arbitrary(this, aim_dir)
        actor.start_firing(this)
        this.next_think_tick= ticks + 5
        --print("  The enemy is in the room.  Aiming at " .. aim_dir)
      else
        actor.stop_firing(this)
        --print("  The enemy is not in the room.")
      end
    end
  end
  if(not this.leaving_place) then
    for name, place in pairs(room.places) do
      if(rects_overlap_aabb(this, place)) then
        local planet= planets[place.name]
        --print("AI " .. this.name .. " entered " .. planet.name)
        local sold_preferred= false
        for tname, tcomm in pairs(this.ship.cargos) do
          if(planet.commodities[tname]) then
            if(planet.commodities[tname].price >= tcomm.bought_price) then
              if(tname == this.preferred_commodity) then
                --print("  " .. this.name .. " is selling its preferred_commodity.")
                sold_preferred= true
              end
              --print("  " .. this.name .. " is selling " .. tcomm.amount .. " of " .. tname .. " for " .. planet.commodities[tname].price)
              fsell(this, planet, tname, tcomm.amount, planet.commodities[tname].price)
            end
          end
        end
        --print("  After selling, " .. this.name .. " has " .. this.credits .. " credits.")
        if(this.ship.cargo_space > 0) then
          if(this.preferred_commodity and not sold_preferred and planet.commodities[this.preferred_commodity]) then
            --print("  " .. this.name .. " is buying its preferred_commodity.")
            ai_buy_comm(this, planet.commodities[this.preferred_commodity], planet)
            --print("  After buying, " .. this.name .. " has " .. this.credits .. " credits.")
          else
            local lowest_name
            local lowest_price
            local lowest_comm
            for pname, pcomm in pairs(planet.commodities) do
              if(not lowest_name or pcomm.price < lowest_price) then
                lowest_name= pname
                lowest_price= pcomm.price
                lowest_comm= pcomm
              end
            end
            if(lowest_name) then
              ai_buy_comm(this, lowest_comm, planet)
              --print("  After buying, " .. this.name .. " has " .. this.credits .. " credits.")
            else
              --print("  " .. this.name .. " couldn't find the lowest priced commodity.")
            end
          end
        end
        this.leaving_place= true
      end
    end
  end
end

function ai_left_room_response(this)
  this.leaving_place= nil
end

function ai_buy_comm(this, comm, planet)
  local ba= this.ship.cargo_space
  if(ba * comm.price > this.credits) then
    ba= math.floor(this.credits / comm.price)
  end
  if(ba <= 0) then
    --print("  " .. this.name .. " can't afford/fit any " .. comm.name)
    return
  end
  --print("  " .. this.name .. " is buying " .. ba .. " of " .. comm.name .. " for " .. comm.price)
  fbuy(this, planet, comm.name, ba, comm.price)
  this.ship.cargos[comm.name].bought_price= comm.price
end

function ai_hit_wall(this, hit_info)
  if(this.cur_rot == 0) then
    if(hit_info.projection.x > 0) then
      -- That means we hit something while moving left.
      this.predisposition= "hop right"
    elseif(hit_info.projection.x < 0) then
      this.predisposition= "hop left"
    end
  elseif(this.cur_rot == 1) then
    if(hit_info.projection.y > 0) then
      this.predisposition= "hop left"
    elseif(hit_info.projection.y < 0) then
      this.predisposition= "hop right"
    end
  elseif(this.cur_rot == 2) then
    if(hit_info.projection.x > 0) then
      this.predisposition= "hop left"
    elseif(hit_info.projection.x < 0) then
      this.predisposition= "hop right"
    end
  elseif(this.cur_rot == 3) then
    if(hit_info.projection.y > 0) then
      this.predisposition= "hop right"
    elseif(hit_info.projection.y < 0) then
      this.predisposition= "hop left"
    end
  end
  --print("AI " .. this.name .. " hit a wall.  Predisposing towards " .. this.predisposition)
  this.next_think_tick= 0
end

function ai_shot_response(this, shot)
  this.enemy= shot.team
  --print(this.name .. " shot, setting enemy to " .. this.enemy)
  this.next_think_tick= 0
end

function create_ai(pos, room, type)
  local ret= create_basic_ava("Geeslaka"..type, "tank1_main_body_sprites.png", room, pos, {math.random(), math.random(), math.random()})
  ret.next_think_tick= math.random(100)
  ret.ai_think_func= basic_ai_think
  ret.ai_room_change_func= ai_left_room_response
  ret.shot_response_func= ai_shot_response
  ret.wall_collide_func= ai_hit_wall
  fill_in_anim_info(ret)
  local pref= math.random(6)
  if(pref == 1) then
    ret.preferred_commodity= "Equipment"
  elseif(pref == 2) then
    ret.preferred_commodity= "Medical"
  elseif(pref == 3) then
    ret.preferred_commodity= "Food"
  elseif(pref == 4) then
    ret.preferred_commodity= "Toads"
  elseif(pref == 5) then
    ret.preferred_commodity= "Tourists"
  -- 6 is for those that don't have a preferred_commodity.
  end
  --print("Created ai at " .. pos.x .. ", " .. pos.y .. " in room " .. room .. " of type " .. type)
  return ret
end

function create_basic_ais()
  for n=1, 0, 1 do
    check_room_active(1, create_ai(sc.Construct("Vec2", 1792, 1792), 1, n))
  end
end

function x_flip_aim(aim)
  local dist_to_mid= (math.pi * .5) - aim
  if aim > math.pi then
    dist_to_mid= (math.pi * 1.5) - aim
  end
  return aim + (dist_to_mid * 2)
end

function no_flip_aim(aim)
  return aim
end

function x_flip_point(p)
  return sc.Construct("Vec2", -p.x, p.y)
end

dir_aim_flip= {left= x_flip_aim, right= no_flip_aim}
-- Due to the loose typing of Lua, no_flip_aim can be reused
dir_offset_flip= {left= x_flip_point, right= no_flip_aim}

ori_to_aim_offset_table= {
  up= 0,
  left= -math.pi*.5,
  down= math.pi,
  right= math.pi*.5
}
ori_to_aim_unoffset_table= {
  up= 0,
  left= math.pi*.5,
  down= math.pi,
  right= -math.pi*.5,
}

function aim_in_range_worker(aim, amin, amax)
  if amin < amax then
    return aim >= amin, aim <= amax
  end
  -- If amin > amax, then the range the aim is supposed to stay inside crosses the 0 to 2pi boundary, so the normal comparison
  -- doesn't work.  So instead, we x flip the aim and the range to get away from the 0 to 2pi boundary.  We also have to swap min
  -- and the max so we don't end up with an inverted range.
  aim= x_flip_aim(aim)
  amin= x_flip_aim(amin)
  amax= x_flip_aim(amax)
  return aim >= amax, aim <= amin, true
end  

function aim_in_range(aim, amin, amax)
  local a, b= aim_in_range_worker(aim, amin, amax)
  return a and b
end

function force_aim_in_range(aim, amin, amax)
  local gte_min, lte_max, swapped= aim_in_range_worker(aim, amin, amax)
  if not swapped then
    if not gte_min then
      return amin
    end
    if not lte_max then
      return amax
    end
    return aim
  end
  if not lte_max then
    return amin
  end
  if not gte_min then
    return amax
  end
  return aim
end

function wrap_aim(aim)
  if aim < 0 then
    aim= aim + (math.pi * 2)
  elseif aim > math.pi * 2 then
    aim= aim - (math.pi * 2)
  end
  return aim
end

function calc_firing_vec(dir, ori, aim, inacc, muzzle_velocity, platform_speed)
  local real_aim= dir_aim_flip[dir](aim) + ori_to_aim_offset_table[ori] + inacc
  local aim_vec= sc.Construct("Vec2", math.cos(real_aim), math.sin(real_aim))
  return (aim_vec * muzzle_velocity) + platform_speed, aim_vec
end

function basic_gun_fire(this, ava, pos, speed, team, croom)
  -- (((math.random(2) - 1) * 2) - 1) will evaluate to either -1 or 1
  -- Proof:
  -- (((1 - 1) * 2) - 1) = ((0 * 2) - 1) = (0 - 1) = -1
  -- (((2 - 1) * 2) - 1) = ((1 * 2) - 1) = (2 - 1) = 1
  local inacc= this.inaccuracy * math.random() * (((math.random(2) - 1) * 2) - 1)
  local avasta= ava.anim_info.status
  local avadir= avasta.dir
  local avaori= avasta.orientation
  local real_pos= pos + rot_point_cen_funcs[ori_to_ind[avaori]](dir_offset_flip[avadir](this.offset))
  local firing_vec, aim_vec= calc_firing_vec(avadir, avaori, this.aim, inacc, this.muzzle_velocity, speed)
  --add_point_to_coords(fireb_coords, fireb_indices, real_pos)
  --add_point_to_coords(fireb_coords, fireb_indices, real_pos + firing_vec)
  create_shot(real_pos, firing_vec, sc.Construct("Vec2", 2, 2), 0, team, "normal", croom)
  if this.ammo_cap ~= -1 then
    ammo_left= ammo_left - 1
  end
  this.aim= this.aim + (this.aim_recoil * (((math.random(2) - 1) * 2) - 1))
  this.aim= wrap_aim(this.aim)
  this.aim= force_aim_in_range(this.aim, this.aim_min, this.aim_max)
  -- FINISHME:  This should be changed to take into account the mass of the tank, but we don't track that yet.
  ava.speed= ava.speed - (aim_vec * this.force_recoil)
  this.next_fire_time= gt_rticks + this.fire_rate
end

function create_basic_ship()
  return {w= 64, h= 32, cargo_space= 100, cargos= {}, fuel= 4, max_fuel= 20,
          primary_guns= {{fire_rate= 1, next_fire_time= 0, ammo_left= -1, ammo_cap= -1, power_draw= 20, aim_recoil= .0872,
                          force_recoil= .01, inaccuracy= .0436, muzzle_velocity= 9, aim_speed= .0872, aim= 0, desired_aim= 0,
                          aim_min= math.pi * 1.75, aim_max= math.pi * .25,
                          offset= sc.Construct("Vec2", 5, -10), firing_func= basic_gun_fire},
                         --[[{fire_rate= 1, next_fire_time= 0, ammo_left= -1, ammo_cap= -1, power_draw= 20, aim_recoil= .0872,
                         force_recoil= .01, inaccuracy= .0436, muzzle_velocity= 9, aim_speed= .0872, aim= math.pi, desired_aim= math.pi,
                         aim_min= math.pi * .75, aim_max= math.pi * 1.25,
                         offset= sc.Construct("Vec2", -10, -10), firing_func= basic_gun_fire}]]
      },
      shield= 50, shield_max= 50, shield_recharge_rate= .0625,
      armor= 50, armor_max= 50, armor_repair_rate= .03125
    }
end

function create_basic_ava(name, image, room, pos, color)
  local ret= {credits= 100, name= name,
    ship= create_basic_ship(),
    jump_max= 35, cur_room= room, cur_rot= 0, cur_planet= "",
    image= load_img_managed(image),
    color= color,
    pos=     pos,
    old_pos= pos,
    g_pos=   pos, -- Filler value.  This will be set correctly the first time update_pos is called.
    g_offset= sc.Construct("Vec2", 0, -32),
    size= sc.Construct("Vec2", 31, 15), -- Slightly smaller than the visual size so that the tank can fit through narrow spaces.
    speed= sc.Construct("Vec2"),
    speed_adjust= sc.Construct("Vec2"),
    general_status= {move_left_down= nil, move_right_down= nil, jump_down= nil},
    anim_info= {status= {dir= "right", orientation= "up", motion= "still", f_num= 0}, frames= {}},
  }
  return ret
end
