--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

function create_shot(pos, speed, size, duration, team, type, room)
--  print("Creating shot:  pos= " .. pos.x .. ", " .. pos.y .. " speed= " .. speed.x .. ", " .. speed.y .. " size= " .. size.x ..
--        ", " .. size.y .. " duration= " .. duration .. " team= " .. team .. " type= " .. type)
  room.shots[#room.shots+1]= {pos= pos, speed= speed, speed_adjust= sc.Construct("Vec2"), size= size, duration= duration, creation_time= gt_ticks, team= team,
                              guide_func= guide_shot_a, draw_func= draw_shot_a, undraw_func= undraw_shot_a, damage_func= shot_damage_a,
                              edge_coll_func= shot_hit_edge_a,
                              wall_collide_func= shot_hit_wall_a
                            }
end

function update_shot(shot)
  -- General function for updating a shot's position after its speed has been calculated and adjusted for any collisions.
  -- Also tracks whether the shot has reached the end of its life.
--  print("Updating shot on team " .. shot.team)
  if(shot.should_be_destroyed) then
    return nil
  end
  if(shot.pos.x < 0 or shot.pos.y < 0 or shot.pos.x > world_max.x or shot.pos.y > world_max.y) then
    shot.edge_coll_func(shot)
  end
  if(shot.duration > 0) then
    shot.duration= shot.duration - 1
    if(shot.duration == 0) then
      return nil
    end
  end
  return true
end

function guide_shot_a(shot)
  -- Simple shot that moves in a straight line.
  -- Since speed was set at creation, that actually means we do nothing.
--  print("Guiding shot on team " .. shot.team)
end

function draw_shot_a(dest, origin, shot)
--  print("Drawing shot on team " .. shot.team)
  dest:SetPrimitiveColor(.75, .5, 0)
  local x, y= coord_transform(origin, shot.pos.x, shot.pos.y)
  dest:DrawPoints(SCUtil.CompileCoords({x, y}), shot.size.x)
end

function undraw_shot_a(dest, origin, shot)
--  print("Undrawing shot on team " .. shot.team)
  local x, y= coord_transform(origin, shot.pos.x, shot.pos.y)
  dest:DrawPoints(SCUtil.CompileCoords({x, y}), shot.size.x)
end

function shot_damage_a(shot, obj)
  obj.was_hit= 3
  local objship= obj.ship
  if obj.shot_response_func then
    obj.shot_response_func(obj, shot)
  end
  if objship then
    if objship.shield > 0 then
      objship.shield= objship.shield - 1
    end
    if objship.shield <= 0 then
      if objship.armor > 0 then
        objship.armor= objship.armor - 1 + objship.shield
        objship.shield= 0
      else
        kill_avatar(obj)
      end
    end
  end
  shot.should_be_destroyed= true
  return 1
end

function shot_hit_edge_a(shot, hit_info)
--  print("Shot on team " .. shot.team .. " hit a edge.")
  shot.should_be_destroyed= true
end

function shot_hit_wall_a(shot, hit_info)
--  print("Shot on team " .. shot.team .. " hit a wall.")
  shot.should_be_destroyed= true
end
