--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

-- Tiles are AABBs, thus x and y refer to the position of the center of the tile.

function create_partially_edged_tile(x, y, t, l, t, r, b)
  --print("    Creating tile at " .. x .. ", " .. y)
  local ret= {type= t,
              pos= sc.Construct("Vec2", x, y),
              size= sc.Construct("Vec2", 16, 16),
              edges= {},
              bounce_mul= 1,
              bounce_add= 0,
              friction_mul= 1,
              friction_add= -1,
            }
  if(l) then ret.edges.left= {a= sc.Construct("Vec2", x-16, y-16), b= sc.Construct("Vec2", x-16, y+16), normal= sc.Construct("Vec2", -1, 0)} end
  if(t) then ret.edges.top= {a= sc.Construct("Vec2", x+16, y-16), b= sc.Construct("Vec2", x-16, y-16), normal= sc.Construct("Vec2", 0, -1)} end
  if(r) then ret.edges.right= {a= sc.Construct("Vec2", x+16, y+16), b= sc.Construct("Vec2", x+16, y-16), normal= sc.Construct("Vec2", 1, 0)} end
  if(b) then ret.edges.bottom= {a= sc.Construct("Vec2", x-16, y+16), b= sc.Construct("Vec2", x+16, y+16), normal= sc.Construct("Vec2", 0, 1)} end
  return ret
end

function create_tile(x, y, t)
  --print("    Creating tile at " .. x .. ", " .. y)
  return {type= t,
              pos= sc.Construct("Vec2", x, y),
              size= sc.Construct("Vec2", 16, 16),
              edges= {left= {a= sc.Construct("Vec2", x-16, y-16), b= sc.Construct("Vec2", x-16, y+16), normal= sc.Construct("Vec2", -1, 0)},
                      bottom= {a= sc.Construct("Vec2", x-16, y+16), b= sc.Construct("Vec2", x+16, y+16), normal= sc.Construct("Vec2", 0, 1)},
                      right= {a= sc.Construct("Vec2", x+16, y+16), b= sc.Construct("Vec2", x+16, y-16), normal= sc.Construct("Vec2", 1, 0)},
                      top= {a= sc.Construct("Vec2", x+16, y-16), b= sc.Construct("Vec2", x-16, y-16), normal= sc.Construct("Vec2", 0, -1)}},
              bounce_mul= 1,
              bounce_add= 0,
              friction_mul= 1,
              friction_add= -1,
            }
end

function tile_get_closest_edge(this, p)
  -- This function will not return the correct edge for cases where edges are of widely varying lengths.
  local closest= {}
  local cdist_sq_sum
  for name,edge in pairs(this.edges) do
    cdist_sq_sum= (edge.a - p):MagnitudeSquared() + (edge.b - p):MagnitudeSquared()
    if(not closest.dist_sq_num or cdist_sq_sum < closest.dist_sq_num) then
      closest.edge= edge
      closest.name= name
      closest.dist_sq_num= cdist_sq_sum
    end
  end
  return closest
end

-- 90 degrees clockwise
function rot_tile1(src)
  local ret= {type= src.type, pos= rot_point1(src.pos, world_max), size= sc.Construct("Vec2", src.size.y, src.size.x),
              bounce_mul= src.bounce_mul, bounce_add= src.bounce_add, friction_mul= src.friction_mul,
              friction_add= src.friction_add}
  ret.edges= {}
  for name, edge in pairs(src.edges) do
    assert(edge.normal)
    ret.edges[name]= {a= rot_point1(edge.a, world_max),
                      b= rot_point1(edge.b, world_max),
                      normal= rot_point_cen1(edge.normal + 1, normal_area) - 1}
  end
  ret.bounce_mul= src.bounce_mul
  ret.bounce_add= src.bounce_add
  ret.friction_mul= src.friction_mul
  ret.friction_add= src.friction_add
  return ret
end

-- 180 degrees
function rot_tile2(src)
  local ret= {type= src.type, pos= rot_point2(src.pos, world_max), size= sc.Construct("Vec2", src.size.x, src.size.y),
              bounce_mul= src.bounce_mul, bounce_add= src.bounce_add, friction_mul= src.friction_mul,
              friction_add= src.friction_add}
  ret.edges= {}
  for name, edge in pairs(src.edges) do
    assert(edge.normal)
    ret.edges[name]= {a= rot_point2(edge.a, world_max),
                      b= rot_point2(edge.b, world_max),
                      normal= rot_point_cen2(edge.normal + 1, normal_area) - 1}
  end
  ret.bounce_mul= src.bounce_mul
  ret.bounce_add= src.bounce_add
  ret.friction_mul= src.friction_mul
  ret.friction_add= src.friction_add
  return ret
end

-- 90 degrees counterclockwise
function rot_tile3(src)
  local ret= {type= src.type, pos= rot_point3(src.pos, world_max), size= sc.Construct("Vec2", src.size.y, src.size.x),
              bounce_mul= src.bounce_mul, bounce_add= src.bounce_add, friction_mul= src.friction_mul,
              friction_add= src.friction_add}
  ret.edges= {}
  for name, edge in pairs(src.edges) do
    assert(edge.normal)
    ret.edges[name]= {a= rot_point3(edge.a, world_max),
                      b= rot_point3(edge.b, world_max),
                      normal= rot_pointcen3(edge.normal + 1, normal_area) - 1}
  end
  ret.bounce_mul= src.bounce_mul
  ret.bounce_add= src.bounce_add
  ret.friction_mul= src.friction_mul
  ret.friction_add= src.friction_add
  return ret
end
